#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test.utils import LunaGatewareTestCase, usb_domain_test_case

from amaranth import Record
from luna.gateware.interface.ulpi import ULPIControlTranslator, ULPIRegisterWindow, ULPIRxEventDecoder, ULPITransmitTranslator

class TestULPIRegisters(LunaGatewareTestCase):
    FRAGMENT_UNDER_TEST = ULPIRegisterWindow

    USB_CLOCK_FREQUENCY = 60e6
    SYNC_CLOCK_FREQUENCY = None

    def initialize_signals(self):
        yield self.dut.ulpi_dir.eq(0)

        yield self.dut.read_request.eq(0)
        yield self.dut.write_request.eq(0)


    @usb_domain_test_case
    def test_idle_behavior(self):
        """ Ensure we apply a NOP whenever we're not actively performing a command. """
        self.assertEqual((yield self.dut.ulpi_data_out), 0)


    @usb_domain_test_case
    def test_register_read(self):
        """ Validates a register read. """

        # Poison the register value with a fail value (0xBD).
        yield self.dut.ulpi_data_in.eq(0xBD)

        # Set up a read request.
        yield self.dut.address.eq(0)
        yield

        # After a read request, we should be busy...
        yield from self.pulse(self.dut.read_request)
        self.assertEqual((yield self.dut.busy), 1)

        # ... and then, since dir is unasserted, we should have a read command.
        yield
        self.assertEqual((yield self.dut.ulpi_data_out), 0b11000000)

        # We should continue to present the command...
        yield from self.advance_cycles(10)
        self.assertEqual((yield self.dut.ulpi_data_out), 0b11000000)
        self.assertEqual((yield self.dut.busy), 1)

        # ... until the host accepts it.
        yield self.dut.ulpi_next.eq(1)
        yield

        # We should then wait for a single bus turnaround cycle before reading.
        yield

        # And then should read whatever value is present.
        yield self.dut.ulpi_data_in.eq(0x07)
        yield
        yield
        self.assertEqual((yield self.dut.read_data), 0x07)

        # Finally, we should return to idle.
        self.assertEqual((yield self.dut.busy), 0)


    @usb_domain_test_case
    def test_interrupted_read(self):
        """ Validates how a register read works when interrupted by a change in DIR. """

        # Set up a read request while DIR is asserted.
        yield self.dut.ulpi_dir.eq(1)
        yield self.dut.address.eq(0)
        yield from self.pulse(self.dut.read_request)

        # We shouldn't try to output anything until DIR is de-asserted.
        yield from self.advance_cycles(1)
        self.assertEqual((yield self.dut.ulpi_out_req), 0)
        yield from self.advance_cycles(10)
        self.assertEqual((yield self.dut.ulpi_out_req), 0)

        # De-assert DIR, and let the platform apply a read command.
        yield self.dut.ulpi_dir.eq(0)
        yield from self.advance_cycles(2)
        self.assertEqual((yield self.dut.ulpi_data_out), 0b11000000)

        # Assert DIR again; interrupting the read. This should bring
        # the platform back to its "waiting for the bus" state.
        yield self.dut.ulpi_dir.eq(1)
        yield from self.advance_cycles(2)
        self.assertEqual((yield self.dut.ulpi_out_req), 0)

        # Clear DIR, and validate that the device starts driving the command again
        yield self.dut.ulpi_dir.eq(0)
        yield from self.advance_cycles(2)
        self.assertEqual((yield self.dut.ulpi_data_out), 0b11000000)

        # Apply NXT so the read can finally continue.
        yield self.dut.ulpi_next.eq(1)
        yield

        # We should then wait for a single bus turnaround cycle before reading.
        yield

        # And then should read whatever value is present.
        yield self.dut.ulpi_data_in.eq(0x07)
        yield
        yield
        self.assertEqual((yield self.dut.read_data), 0x07)

        # Finally, we should return to idle.
        self.assertEqual((yield self.dut.busy), 0)


    @usb_domain_test_case
    def test_register_write(self):

        # Set up a write request.
        yield self.dut.address.eq(0b10)
        yield self.dut.write_data.eq(0xBC)
        yield

        # Starting the request should make us busy.
        yield from self.pulse(self.dut.write_request)
        self.assertEqual((yield self.dut.busy), 1)

        # ... and then, since dir is unasserted, we should have a write command.
        yield
        self.assertEqual((yield self.dut.ulpi_data_out), 0b10000010)

        # We should continue to present the command...
        yield from self.advance_cycles(10)
        self.assertEqual((yield self.dut.ulpi_data_out), 0b10000010)
        self.assertEqual((yield self.dut.busy), 1)

        # ... until the host accepts it.
        yield self.dut.ulpi_next.eq(1)
        yield

        # We should then present the data to be written...
        yield self.dut.ulpi_next.eq(0)
        yield
        self.assertEqual((yield self.dut.ulpi_data_out), 0xBC)

        # ... and continue doing so until the host accepts it...
        yield from self.advance_cycles(10)
        self.assertEqual((yield self.dut.ulpi_data_out), 0xBC)

        yield self.dut.ulpi_next.eq(1)
        yield from self.advance_cycles(2)

        # ... at which point stop should be asserted for one cycle.
        self.assertEqual((yield self.dut.ulpi_stop), 1)
        yield

        # Finally, we should go idle.
        self.assertEqual((yield self.dut.ulpi_stop), 0)
        self.assertEqual((yield self.dut.busy), 0)


class ULPIRxEventDecoderTest(LunaGatewareTestCase):

    USB_CLOCK_FREQUENCY = 60e6
    SYNC_CLOCK_FREQUENCY = None

    def instantiate_dut(self):

        self.ulpi = Record([
            ("dir", [
                ("i", 1),
            ]),
            ("nxt", [
                ("i", 1),
            ]),
            ("data", [
                ("i", 8),
            ])
        ])

        return ULPIRxEventDecoder(ulpi_bus=self.ulpi)


    def initialize_signals(self):
        yield self.ulpi.dir.i.eq(0)
        yield self.ulpi.nxt.i.eq(0)
        yield self.ulpi.data.i.eq(0)
        yield self.dut.register_operation_in_progress.eq(0)


    @usb_domain_test_case
    def test_decode(self):

        # Provide a test value.
        yield self.ulpi.data.i.eq(0xAB)

        # First, set DIR and NXT at the same time, and verify that we
        # don't register an RxEvent.
        yield self.ulpi.dir.i.eq(1)
        yield self.ulpi.nxt.i.eq(1)

        yield from self.advance_cycles(5)
        self.assertEqual((yield self.dut.last_rx_command), 0x00)

        # Nothing should change when we drop DIR and NXT.
        yield self.ulpi.dir.i.eq(0)
        yield self.ulpi.nxt.i.eq(0)
        yield
        self.assertEqual((yield self.dut.last_rx_command), 0x00)


        # Setting DIR but not NXT should trigger an RxEvent; but not
        # until one cycle of "bus turnaround" has passed.
        yield self.ulpi.dir.i.eq(1)

        yield self.ulpi.data.i.eq(0x12)
        yield
        self.assertEqual((yield self.dut.last_rx_command), 0x00)

        yield self.ulpi.data.i.eq(0b00011110)
        yield from self.advance_cycles(2)

        self.assertEqual((yield self.dut.last_rx_command), 0b00011110)

        # Validate that we're decoding this RxCommand correctly.
        self.assertEqual((yield self.dut.line_state),     0b10)
        self.assertEqual((yield self.dut.vbus_valid),        1)
        self.assertEqual((yield self.dut.rx_active),         1)
        self.assertEqual((yield self.dut.rx_error),          0)
        self.assertEqual((yield self.dut.host_disconnect),   0)


class ControlTranslatorTest(LunaGatewareTestCase):

    USB_CLOCK_FREQUENCY = 60e6
    SYNC_CLOCK_FREQUENCY = None

    def instantiate_dut(self):
        self.reg_window = ULPIRegisterWindow()
        return ULPIControlTranslator(register_window=self.reg_window, own_register_window=True)


    def initialize_signals(self):
        dut = self.dut

        # Initialize our register signals to their default values.
        yield dut.xcvr_select.eq(1)
        yield dut.dm_pulldown.eq(1)
        yield dut.dp_pulldown.eq(1)
        yield dut.use_external_vbus_indicator.eq(0)
        yield dut.bus_idle.eq(1)


    @usb_domain_test_case
    def test_multiwrite_behavior(self):

        # Give our initialization some time to settle,
        # and verify that we haven't initiated anyting in that interim.
        yield from self.advance_cycles(10)
        self.assertEqual((yield self.reg_window.write_request), 0)

        # Change signals that span two registers.
        yield self.dut.op_mode.eq(0b11)
        yield self.dut.dp_pulldown.eq(0)
        yield self.dut.dm_pulldown.eq(0)
        yield
        yield

        # Once we've changed these, we should start trying to apply
        # our new value to the function control register.
        self.assertEqual((yield self.reg_window.address),      0x04)
        self.assertEqual((yield self.reg_window.write_data),   0b01011001)

        # which should occur until the data and address are accepted.
        yield self.reg_window.ulpi_next.eq(1)
        yield from self.wait_until(self.reg_window.done, timeout=10)
        yield
        yield

        # We should then experience a write to the function control register.
        self.assertEqual((yield self.reg_window.address),      0x0A)
        self.assertEqual((yield self.reg_window.write_data),   0b00000000)

        # Wait for that action to complete..
        yield self.reg_window.ulpi_next.eq(1)
        yield from self.wait_until(self.reg_window.done, timeout=10)
        yield
        yield

        # After which we shouldn't be trying to write anything at all.
        self.assertEqual((yield self.reg_window.address),       0)
        self.assertEqual((yield self.reg_window.write_data),    0)
        self.assertEqual((yield self.reg_window.write_request), 0)


class ULPITransmitTranslatorTest(LunaGatewareTestCase):
    USB_CLOCK_FREQUENCY=60e6
    SYNC_CLOCK_FREQUENCY=None

    FRAGMENT_UNDER_TEST = ULPITransmitTranslator

    def initialize_signals(self):
        yield self.dut.bus_idle.eq(1)

    @usb_domain_test_case
    def test_simple_transmit(self):
        dut = self.dut

        # We shouldn't try to transmit until we have a transmit request.
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.ulpi_out_req), 0)

        # Present a simple SOF PID.
        yield dut.tx_valid.eq(1)
        yield dut.tx_data.eq(0xA5)
        yield

        # Our PID should have been translated into a transmit request, with
        # our PID in the lower nibble.
        self.assertEqual((yield dut.ulpi_data_out), 0b01000101)
        self.assertEqual((yield dut.tx_ready),      0)
        self.assertEqual((yield dut.ulpi_stp),      0)
        yield
        self.assertEqual((yield dut.ulpi_out_req),  1)

        # Our PID should remain there until we indicate we're ready.
        self.advance_cycles(10)
        self.assertEqual((yield dut.ulpi_data_out), 0b01000101)

        # Once we're ready, we should accept the data from the link and continue.
        yield dut.ulpi_nxt.eq(1)
        yield
        self.assertEqual((yield dut.tx_ready),      1)
        yield dut.tx_data.eq(0x11)
        yield

        # At which point we should present the relevant data directly.
        yield
        self.assertEqual((yield dut.ulpi_data_out), 0x11)

        # Finally, once we stop our transaction...
        yield dut.tx_valid.eq(0)
        yield

        # ... we should get a cycle of STP.
        self.assertEqual((yield dut.ulpi_data_out), 0)
        self.assertEqual((yield dut.ulpi_stp),      1)

        # ... followed by idle.
        yield
        self.assertEqual((yield dut.ulpi_stp),      0)


    @usb_domain_test_case
    def test_handshake(self):
        dut = self.dut

        # Present a simple ACK PID.
        yield dut.tx_valid.eq(1)
        yield dut.tx_data.eq(0b11010010)
        yield

        # Our PID should have been translated into a transmit request, with
        # our PID in the lower nibble.
        self.assertEqual((yield dut.ulpi_data_out), 0b01000010)
        self.assertEqual((yield dut.tx_ready),      0)
        self.assertEqual((yield dut.ulpi_stp),      0)

        # Once the PHY accepts the data, it'll assert NXT.
        yield dut.ulpi_nxt.eq(1)
        yield
        self.assertEqual((yield dut.ulpi_out_req),  1)

        # ... which will trigger the transmitter to drop tx_valid.
        yield dut.tx_valid.eq(0)

        # ... we should get a cycle of STP.
        yield
        #self.assertEqual((yield dut.ulpi_data_out), 0)
        #self.assertEqual((yield dut.ulpi_stp),      1)

        # ... followed by idle.
        yield
        self.assertEqual((yield dut.ulpi_stp),      0)




#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Gateware that handles USB bus resets & speed detection. """

from amaranth              import *

from .                     import USBSpeed
from ...interface.utmi     import UTMITransmitInterface, UTMIOperatingMode, UTMITerminationSelect


def _generate_wide_incrementer(m, platform, adder_input):
    """ Attempts to create an optimal wide-incrementer for counters.

    Yosys on certain platforms (ice40 UltraPlus) doesn't currently use hardware resources
    effectively for wide adders. We'll manually instantiate the relevant resources
    to get rid of an 18-bit carry chain; avoiding a long critical path.

    Parameters:
        platform    -- The platform we're working with.
        adder_input -- The input to our incrementer.
    """

    # If this isn't an iCE40 UltraPlus, let Yosys do its thing.
    if (not platform) or not platform.device.startswith('iCE40UP'):
        return adder_input + 1

    # Otherwise, we'll create a DSP adder itself.
    output = Signal.like(adder_input)
    m.submodules += Instance('SB_MAC16',

        # Hook up our inputs and outputs.
        # A = upper bits of input; B = lower bits of input
        i_A      = adder_input[16:],
        i_B      = adder_input[0:16],
        o_O      = output,

        p_TOPADDSUB_UPPERINPUT =0b1,  # Use as a normal adder
        p_TOPADDSUB_CARRYSELECT=0b11, # Connect our top and bottom adders together.
        p_BOTADDSUB_UPPERINPUT =0b1,  # Use as a normal adder.
        p_BOTADDSUB_CARRYSELECT=0b01  # Always increment.
    )

    return output



class USBResetSequencer(Elaboratable):
    """ Gateware that detects reset signaling on the USB bus.

    Attributes
    ----------
    low_speed_only: Signal(), input
        If set, the device will be forced to operate as a low-speed device.
    prevent_high_speed: Signal(), input
        If set, the device will be prohibited from entering high-speed states; and will thus
        act like it's a full speed device (low_speed_only = 0).
    bus_busy: Signal(), input
        Hold-off signal that indicates that driving the bus should be delayed.
    vbus_connected: Signal(), input
        Indicates that the device is connected to VBUS. When this is de-asserted, the device will
        be held in perpetual bus reset, and reset handshaking will be disabled.
    line_state: Signal(2), input
        The UTMI linestate signals; used to read the current state of the USB D+ and D- lines.
    disconnect: Signal(), input
        If set, the device will be switched into non-driving operating mode to force a host disconnect.

    bus_reset: Signal(), output
        Strobe; pulses high for one cycle when a bus reset is detected. This signal indicates that the
         device should return to unaddressed, unconfigured, and should not longer be in High Speed mode.
    suspended: Signal(), output
        Held high while the USB device should be in suspend. This technically indicates that the device should
        drop down to consuming suspend current (<= 2.5mA), but very few devices are compliant with this requirement.
        Either way, a polite device might reduce its power consumption while in suspend.

    current_speed: Signal(2), output
        A USBSpeed value that indicates the current operating speed. Used both to drive our device's
        knowledge of operating speed and to drive our PHY's speed selection.
    operating_mode: Signal(2), output
        The current UTMI operating mode. Used to select whether we're driving the USB bus directly;
        or whether we're letting the PHY handle NRZI/bit-stuffing.
    termination_select: Signal(), output, default=1
        Determines the bus termination mode. In LS/FS, this determines the presence of our presence-detect
        pull-up. In HS mode, this determines whether the USB high-speed termination is present (0), or
        whether we're in chirp mode (1).

    tx: UTMITransmitInterface, output stream
                     -- Our UTMI transmit interface; used to drive chirp signaling onto the bus.
    """

    # Constants for our line states at various speeds.
    _LINE_STATE_SE0       = 0b00
    _LINE_STATE_SQUELCH   = 0b00
    _LINE_STATE_FS_HS_K   = 0b10
    _LINE_STATE_FS_HS_J   = 0b01
    _LINE_STATE_LS_K      = 0b01
    _LINE_STATE_LS_J      = 0b10

    # Reset time constants.
    # Eventually, if we support clocks other than 60MHz (48 MHz)?
    # We should provide the ability to scale these.
    _CYCLES_500_NANOSECONDS    = 30
    _CYCLES_1_MICROSECOND      = _CYCLES_500_NANOSECONDS  * 2
    _CYCLES_2P5_MICROSECONDS   = _CYCLES_500_NANOSECONDS  * 5
    _CYCLES_5_MICROSECONDS     = _CYCLES_1_MICROSECOND    * 5
    _CYCLES_200_MICROSECONDS   = _CYCLES_1_MICROSECOND    * 200
    _CYCLES_1_MILLISECONDS     = _CYCLES_1_MICROSECOND    * 1000
    _CYCLES_2_MILLISECONDS     = _CYCLES_1_MILLISECONDS   * 2
    _CYCLES_2P5_MILLISECONDS   = _CYCLES_2P5_MICROSECONDS * 1000
    _CYCLES_3_MILLISECONDS     = _CYCLES_1_MILLISECONDS   * 3


    def __init__(self):

        #
        # I/O port
        #
        self.low_speed_only     = Signal()
        self.full_speed_only    = Signal()

        self.bus_busy           = Signal()
        self.vbus_connected     = Signal()
        self.line_state         = Signal(2)

        self.disconnect         = Signal()

        self.bus_reset          = Signal()
        self.suspended          = Signal()

        self.current_speed      = Signal(2, init=USBSpeed.FULL)
        self.operating_mode     = Signal(2, init=UTMIOperatingMode.NORMAL)
        self.termination_select = Signal(1, init=1)

        self.tx                 = UTMITransmitInterface()


    def elaborate(self, platform):
        m = Module()

        if hasattr(platform, 'ignore_phy_vbus') and platform.ignore_phy_vbus:
            self.vbus_connected = Const(1)

        # Event timer: keeps track of the timing of each of the individual event phases.
        timer = Signal(range(0, self._CYCLES_3_MILLISECONDS + 1))

        # Line state timer: keeps track of how long we've seen a line-state of interest;
        # other than a reset SE0. Used to track chirp and idle times.
        line_state_time = Signal(range(0, self._CYCLES_3_MILLISECONDS + 1))

        # Valid pairs: keeps track of how make Chirp K / Chirp J sequences we've
        # seen, thus far.
        valid_pairs = Signal(range(0, 4))

        # Tracks whether we were at high speed when we entered a suspend state.
        was_hs_pre_suspend = Signal()

        # By default, always count forward in time.
        # We'll reset the timer below when appropriate.
        m.d.usb += timer.eq(_generate_wide_incrementer(m, platform, timer))
        m.d.usb += line_state_time.eq(_generate_wide_incrementer(m, platform, line_state_time))

        # Signal that indicates when the bus is idle.
        # Our bus's IDLE condition depends on our active speed.
        bus_idle = Signal()

        # High speed busses present SE0 (which we see as SQUELCH'd) when idle [USB2.0: 7.1.1.3].
        with m.If(self.current_speed == USBSpeed.HIGH):
            m.d.comb += bus_idle.eq(self.line_state == self._LINE_STATE_SQUELCH)

        # Full and low-speed busses see a 'J' state when idle, due to the device pull-up resistors.
        # (The line_state values for these are flipped between speeds.) [USB2.0: 7.1.7.4.1; USB2.0: Table 7-2].
        with m.Elif(self.current_speed == USBSpeed.FULL):
            m.d.comb += bus_idle.eq(self.line_state == self._LINE_STATE_FS_HS_J)
        with m.Else():
            m.d.comb += bus_idle.eq(self.line_state == self._LINE_STATE_LS_J)


        #
        # Core reset sequences.
        #
        with m.FSM(domain='usb'):

            # INITIALIZE -- we're immediately post-reset; we'll perform some minor setup
            with m.State('INITIALIZE'):

                # If we're working in low-speed mode, configure the PHY accordingly.
                with m.If(self.low_speed_only):
                    m.d.usb += self.current_speed.eq(USBSpeed.LOW)

                m.next = 'LS_FS_NON_RESET'
                m.d.usb += [
                    timer.eq(0),
                    line_state_time.eq(0)
                ]

            # LS_FS_NON_RESET -- we're currently operating at LS/FS and waiting for a reset;
            # the device could be active or inactive, but we haven't yet seen a reset condition.
            with m.State('LS_FS_NON_RESET'):

                # If we're seeing a state other than SE0 (D+ / D- at zero), this isn't yet a
                # potential reset. Keep our timer at zero.
                with m.If(self.line_state != self._LINE_STATE_SE0):
                    m.d.usb += timer.eq(0)
                    # Enter forced disconnect when self.disconnect is high.
                    with m.If(self.disconnect):
                        m.next = "DISCONNECT"

                # If VBUS isn't connected, don't go through the whole reset process;
                # but also consider ourselves permanently in reset. This ensures we
                # don't progress through the reset FSM; but also ensures the device
                # state starts fresh with each plug.
                with m.If(~self.vbus_connected):
                    m.d.usb  += timer.eq(0)
                    m.d.comb += self.bus_reset.eq(1)

                # If we see an SE0 for >2.5uS; < 3ms, this a bus reset.
                # We'll trigger a reset after 5uS; providing a little bit of timing flexibility.
                # [USB2.0: 7.1.7.5; ULPI 3.8.5.1].
                with m.If(timer == self._CYCLES_5_MICROSECONDS):
                    m.d.comb += self.bus_reset.eq(1)

                    # If we're okay to run in high speed, we'll try to perform a high-speed detect.
                    with m.If(~self.low_speed_only & ~self.full_speed_only):
                        m.next = 'START_HS_DETECTION'


                # If we're seeing a state other than IDLE, clear our suspend timer.
                with m.If(~bus_idle):
                    m.d.usb += line_state_time.eq(0)

                # If we see 3ms of consecutive line idle, we're being put into USB suspend.
                # We'll enter our suspended state, directly. [USB2.0: 7.1.7.6]
                with m.If(line_state_time == self._CYCLES_3_MILLISECONDS):
                    m.d.usb += was_hs_pre_suspend.eq(0)
                    m.next = 'SUSPENDED'




            # HS_NON_RESET -- we're currently operating at high speed and waiting for a reset or
            # suspend; the device could be active or inactive.
            with m.State('HS_NON_RESET'):

                # If we're seeing a state other than SE0 (D+ / D- at zero), this isn't yet a
                # potential reset. Keep our timer at zero.
                with m.If(self.line_state != self._LINE_STATE_SE0):
                    m.d.usb += timer.eq(0)
                    # Enter forced disconnect when self.disconnect is high.
                    with m.If(self.disconnect):
                        m.next = "DISCONNECT"

                # If VBUS isn't connected, our device/host relationship is effectively
                # a blank state. We'll want to present our detection pull-up to the host,
                # so we'll drop out of high speed.
                with m.If(~self.vbus_connected):
                    m.d.comb += self.bus_reset.eq(1)
                    m.next = 'IS_LOW_OR_FULL_SPEED'


                # High speed signaling presents IDLE and RESET the same way: with the host
                # driving SE0; and us seeing SQUELCH. [USB2.0: 7.1.1.3; USB2.0: 7.1.7.6].
                # Either way, our next step is the same: we'll drop down to full-speed. [USB2.0: 7.1.7.6]
                # Afterwards, we'll take steps to differentiate a reset from a suspend.
                with m.If(timer == self._CYCLES_3_MILLISECONDS):
                    m.d.usb += [
                        timer                    .eq(0),

                        self.current_speed       .eq(USBSpeed.FULL),
                        self.operating_mode      .eq(UTMIOperatingMode.NORMAL),
                        self.termination_select  .eq(UTMITerminationSelect.LS_FS_NORMAL),
                    ]
                    m.next = 'DETECT_HS_SUSPEND'


                # If we see full-speed-only or low-speed-only being driven, switch
                # back to our LS/FS mode.
                with m.If(self.full_speed_only | self.low_speed_only):
                    m.next = 'IS_LOW_OR_FULL_SPEED'


            # START_HS_DETECTION -- entry state for high-speed detection
            with m.State('START_HS_DETECTION'):
                m.d.usb += [
                    timer                    .eq(0),

                    # Switch into High-speed chirp mode. Note that we'll need to leave our
                    # terminations set to '1' until we're sure this is a high-speed host;
                    # or the host will see our pull-up removal as a disconnect.
                    self.current_speed       .eq(USBSpeed.HIGH),
                    self.operating_mode      .eq(UTMIOperatingMode.CHIRP),
                    self.termination_select  .eq(UTMITerminationSelect.HS_CHIRP)
                ]
                m.next = 'PREPARE_FOR_CHIRP_0'


            # PREPARE_FOR_CHIRP_0 / PREPARE_FOR_CHIRP_1-- wait states; in which we give the PHY
            # time to the mode we'll need to drive our high-speed chirp.
            with m.State('PREPARE_FOR_CHIRP_0'):
                with m.If(~self.bus_busy):
                    m.next = 'PREPARE_FOR_CHIRP_1'

            with m.State('PREPARE_FOR_CHIRP_1'):
                with m.If(~self.bus_busy):
                    m.next = 'DEVICE_CHIRP'


            # DEVICE_CHIRP -- the device produces a 'chirp' K, which advertises to the host that
            # we're high speed capable. We'll provide that chirp K for around ~2ms. [USB2.0: 7.1.7.5]
            with m.State('DEVICE_CHIRP'):

                # Transmit a constant stream of 0's, which in this mode is a Chirp K.
                # Note that we don't need to check 'ready', as we care about the length
                # of time, rather than the number of bits.
                m.d.comb += [
                    self.tx.valid  .eq(1),
                    self.tx.data   .eq(0)
                ]

                # Once 2ms have passed, we can stop our chirp, and begin waiting for the
                # hosts's response. We'll wait for Ready to be asserted to do so, to ensure
                # we don't change our values in the middle of a bit.
                with m.If((timer == self._CYCLES_2_MILLISECONDS)):
                    m.d.usb += [
                        timer        .eq(0),
                        valid_pairs  .eq(0)
                    ]
                    m.next = 'AWAIT_HOST_K'


            # AWAIT_HOST_K -- we've now completed the device chirp; and are waiting to see if the host
            # will respond with an alternating sequence of K's and J's.
            with m.State('AWAIT_HOST_K'):

                # If we don't see our response within 2.5ms, this isn't a compliant HS host. [USB2.0: 7.1.7.5].
                # This is thus a full-speed host, and we'll act as a full-speed device.
                with m.If(timer == self._CYCLES_2P5_MILLISECONDS):
                    m.next = 'IS_LOW_OR_FULL_SPEED'

                # Once we've seen our K, we're good to start observing J/K toggles.
                with m.If(self.line_state == self._LINE_STATE_FS_HS_K):
                    m.next = 'IN_HOST_K'
                    m.d.usb += line_state_time.eq(0)


            # IN_HOST_K: we're seeing a host Chirp K as part of our handshake; we'll
            # time it and see how long it lasts
            with m.State('IN_HOST_K'):

                # If we've exceeded our minimum chirp time, consider this a valid pattern
                # bit, # and advance in the pattern.
                with m.If(line_state_time == self._CYCLES_2P5_MICROSECONDS):
                    m.next = 'AWAIT_HOST_J'

                # If our input has become something other than a K, then
                # we haven't finished our sequence. We'll go back to expecting a K.
                with m.If(self.line_state != self._LINE_STATE_FS_HS_K):
                    m.next = 'AWAIT_HOST_K'

                # Time out if we exceed our maximum allowed duration.
                with m.If(timer == self._CYCLES_2P5_MILLISECONDS):
                    m.next = 'IS_LOW_OR_FULL_SPEED'


            # AWAIT_HOST_J -- we're waiting for the next Chirp J in the host chirp sequence
            with m.State('AWAIT_HOST_J'):

                # If we've exceeded our maximum wait, this isn't a high speed host.
                with m.If(timer == self._CYCLES_2P5_MILLISECONDS):
                    m.next = 'IS_LOW_OR_FULL_SPEED'

                # Once we've seen our J, start timing its duration.
                with m.If(self.line_state == self._LINE_STATE_FS_HS_J):
                    m.next = 'IN_HOST_J'
                    m.d.usb += line_state_time.eq(0)


            # IN_HOST_J: we're seeing a host Chirp K as part of our handshake; we'll
            # time it and see how long it lasts
            with m.State('IN_HOST_J'):

                # If we've exceeded our minimum chirp time, consider this a valid pattern
                # bit, and advance in the pattern.
                with m.If(line_state_time == self._CYCLES_2P5_MICROSECONDS):

                    # If this would complete our third pair, this completes a handshake,
                    # and we've identified a high speed host!
                    with m.If(valid_pairs == 2):
                        m.next = 'IS_HIGH_SPEED'

                    # Otherwise, count the pair as valid, and wait for the next K.
                    with m.Else():
                        m.d.usb += valid_pairs.eq(valid_pairs + 1)
                        m.next = 'AWAIT_HOST_K'

                # If our input has become something other than a K, then
                # we haven't finished our sequence. We'll go back to expecting a K.
                with m.If(self.line_state != self._LINE_STATE_FS_HS_J):
                    m.next = 'AWAIT_HOST_J'

                # Time out if we exceed our maximum allowed duration.
                with m.If(timer == self._CYCLES_2P5_MILLISECONDS):
                    m.next = 'IS_LOW_OR_FULL_SPEED'


            # IS_HIGH_SPEED -- we've completed a high speed handshake, and are ready to
            # switch to high speed signaling
            with m.State('IS_HIGH_SPEED'):

                # Switch to high speed.
                m.d.usb += [
                    timer                    .eq(0),
                    line_state_time          .eq(0),

                    self.current_speed       .eq(USBSpeed.HIGH),
                    self.operating_mode      .eq(UTMIOperatingMode.NORMAL),
                    self.termination_select  .eq(UTMITerminationSelect.HS_NORMAL)
                ]

                m.next = 'HS_NON_RESET'


            # IS_LOW_OR_FULL_SPEED -- we've decided the device is low/full speed (typically
            # because it didn't) complete our high-speed handshake; set it up accordingly.
            with m.State('IS_LOW_OR_FULL_SPEED'):
                m.d.usb += [
                    self.operating_mode      .eq(UTMIOperatingMode.NORMAL),
                    self.termination_select  .eq(UTMITerminationSelect.LS_FS_NORMAL)
                ]

                # If we're operating in low-speed only, drop down to low speed.
                with m.If(self.low_speed_only):
                    m.d.usb += self.current_speed.eq(USBSpeed.LOW),

                # Otherwise, drop down to full speed.
                with m.Else():
                    m.d.usb += self.current_speed.eq(USBSpeed.FULL)

                # Once we know that our reset is complete, move back to our normal, non-reset state.
                with m.If(self.line_state != self._LINE_STATE_SE0):
                    m.next = 'LS_FS_NON_RESET'
                    m.d.usb += [
                        timer.eq(0),
                        line_state_time.eq(0)
                    ]


            # DETECT_HS_SUSPEND -- we were operating at high speed, and just detected an event
            # which is either a reset or a suspend event; we'll now detect which.
            with m.State('DETECT_HS_SUSPEND'):

                # We've just switch from HS signaling to FS signaling.
                # We'll wait a little while for the bus to settle, and then
                # check to see if it's settled to FS idle; or if we still see SE0.
                with m.If(timer == self._CYCLES_200_MICROSECONDS):
                    m.d.usb += timer.eq(0)

                    # If we've resume IDLE, this is suspend. Move to HS suspend.
                    with m.If(self.line_state == self._LINE_STATE_FS_HS_J):
                        m.d.usb += was_hs_pre_suspend.eq(1)
                        m.next = 'SUSPENDED'

                    # Otherwise, this is a reset (or, if K/SE1, we're very confused, and
                    # should re-initialize anyway). Move to the HS reset detect sequence.
                    with m.Else():
                        m.d.comb += self.bus_reset.eq(1)
                        m.next = 'START_HS_DETECTION'


            # SUSPEND -- our device has entered USB suspend; we'll now wait for either a
            # resume or a reset
            with m.State('SUSPENDED'):
                m.d.comb += self.suspended.eq(1)

                # If we see a K state, then we're being resumed.
                is_ls_k = self.low_speed_only  & (self.line_state == self._LINE_STATE_LS_K)
                is_fs_k = ~self.low_speed_only & (self.line_state == self._LINE_STATE_FS_HS_K)
                with m.If(is_ls_k | is_fs_k):
                    m.d.usb  += timer.eq(0)

                    # If we were in high-speed pre-suspend, then resume being in HS.
                    with m.If(was_hs_pre_suspend):
                        m.next = 'IS_HIGH_SPEED'

                    # Otherwise, just resume.
                    with m.Else():
                        m.next = 'LS_FS_NON_RESET'
                        m.d.usb += [
                            timer.eq(0),
                            line_state_time.eq(0)
                        ]


                # If this isn't an SE0, we're not receiving a reset request.
                # Keep our reset counter at zero.
                with m.If(self.line_state != self._LINE_STATE_SE0):
                    m.d.usb += timer.eq(0)


                # If we see an SE0 for > 2.5uS, this is a reset request. [USB 2.0: 7.1.7.5]
                # We'll handle it directly from suspend.
                with m.If(timer == self._CYCLES_2P5_MICROSECONDS):
                    m.d.comb += self.bus_reset.eq(1)
                    m.d.usb  += timer.eq(0)

                    # If we're limited to LS or FS, move to the appropriate state.
                    with m.If(self.low_speed_only | self.full_speed_only):
                        m.next = 'LS_FS_NON_RESET'
                        m.d.usb += [
                            timer.eq(0),
                            line_state_time.eq(0)
                        ]

                    # Otherwise, this could be a high-speed device; enter its reset.
                    with m.Else():
                        m.next = 'START_HS_DETECTION'


            # DISCONNECT -- our device has entered a forced USB disconnect; hold the device in
            # NON_DRIVING operating mode for Tddis=0.25us and wait for self.disconnect to go low.
            with m.State('DISCONNECT'):
                m.d.usb += self.operating_mode.eq(UTMIOperatingMode.NON_DRIVING)

                # A disconnect condition is indicated if the host or hub is not driving the data lines and an
                # SE0 persists on a downstream facing port for more than Tddis.
                # [USB2.0: 7.1.7.3].
                tddis = Signal()
                with m.If(timer == self._CYCLES_2P5_MICROSECONDS):
                    m.d.usb += tddis.eq(1)

                # Exit DISCONNECT once the Tddis timer has expired and self.disconnect is low.
                with m.If((~self.disconnect) & tddis):
                    m.d.usb += [
                        tddis.eq(0),
                        self.current_speed.eq(USBSpeed.FULL),
                        self.operating_mode.eq(UTMIOperatingMode.NORMAL),
                        self.termination_select.eq(1),
                    ]
                    m.next = 'INITIALIZE'

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Header Packet Rx-handling gateware. """

from amaranth                      import *

from usb_protocol.types.superspeed import LinkCommand

from .header                       import HeaderPacket, HeaderQueue
from .crc                          import compute_usb_crc5, HeaderPacketCRC
from .command                      import LinkCommandGenerator
from ..physical.coding             import SHP, EPF, stream_matches_symbols
from ...stream                     import USBRawSuperSpeedStream


class RawHeaderPacketReceiver(Elaboratable):
    """ Class that monitors the USB bus for Header Packet, and receives them.

    This class performs the validations required at the link layer of the USB specification;
    which include checking the CRC-5 and CRC-16 embedded within the header packet.


    Attributes
    ----------
    sink: USBRawSuperSpeedStream(), input (monitor only)
        Stream that the USB data to be monitored.
    packet: HeaderPacket(), output
        The de-serialized form of our header packet.

    new_packet: Signal(), output
        Strobe; indicates that a new, valid header packet has been received. The new
        packet is now available on :attr:``packet``.
    bad_packet: Signal(), output
        Strobe; indicates that a corrupted, invalid header packet has been received.
        :attr:``packet`` has not been updated.

    expected_sequence: Signal(3), input
        Indicates the next expected sequence number; used to validate the received packet.

    """

    def __init__(self):

        #
        # I/O port
        #
        self.sink              = USBRawSuperSpeedStream()

        # Header packet output.
        self.packet            = HeaderPacket()

        # State indications.
        self.new_packet        = Signal()
        self.bad_packet        = Signal()

        # Sequence tracking.
        self.expected_sequence = Signal(3)
        self.bad_sequence      = Signal()


    def elaborate(self, platform):
        m = Module()

        sink   = self.sink

        # Store our header packet in progress; which we'll output only once it's been validated.
        packet = HeaderPacket()

        # Cache our expected CRC5, so we can pipeline generation and comparison.
        expected_crc5 = Signal(5)

        # Keep our "new packet" signal de-asserted unless asserted explicitly.
        m.d.ss += self.new_packet.eq(0)

        #
        # CRC-16 Generator
        #
        m.submodules.crc16 = crc16 = HeaderPacketCRC()
        m.d.comb += crc16.data_input.eq(sink.data),


        #
        # Receiver Sequencing
        #
        with m.FSM(domain="ss"):

            # WAIT_FOR_HPSTART -- we're currently waiting for HPSTART framing, which indicates
            # that the following 16 symbols (4 words) will be a header packet.
            with m.State("WAIT_FOR_HPSTART"):

                # Don't start our CRC until we're past our HPSTART header.
                m.d.comb += crc16.clear.eq(1)

                is_hpstart = stream_matches_symbols(sink, SHP, SHP, SHP, EPF)
                with m.If(is_hpstart):
                    m.next = "RECEIVE_DW0"

            # RECEIVE_DWn -- the first three words of our header packet are data words meant form
            # the protocol layer; we'll receive them so we can pass them on to the protocol layer.
            for n in range(3):
                with m.State(f"RECEIVE_DW{n}"):

                    with m.If(sink.valid):
                        m.d.comb += crc16.advance_crc.eq(1)
                        m.d.ss += packet[f'dw{n}'].eq(sink.data)
                        m.next = f"RECEIVE_DW{n+1}"

            # RECEIVE_DW3 -- we'll receive and parse our final data word, which contains the fields
            # relevant to the link layer.
            with m.State("RECEIVE_DW3"):

                with m.If(sink.valid):
                    m.d.ss += [
                        # Collect the fields from the DW...
                        packet.crc16            .eq(sink.data[ 0:16]),
                        packet.sequence_number  .eq(sink.data[16:19]),
                        packet.dw3_reserved     .eq(sink.data[19:22]),
                        packet.hub_depth        .eq(sink.data[22:25]),
                        packet.delayed          .eq(sink.data[25]),
                        packet.deferred         .eq(sink.data[26]),
                        packet.crc5             .eq(sink.data[27:32]),

                        # ... and pipeline a CRC of the to the link control word.
                        expected_crc5           .eq(compute_usb_crc5(sink.data[16:27]))
                    ]

                    m.next = "CHECK_PACKET"

            # CHECK_PACKET -- we've now received our full packet; we'll check it for validity.
            with m.State("CHECK_PACKET"):

                # A minor error occurs if if one of our CRCs mismatches; in which case the link can
                # continue after sending an LBAD link command. [USB3.2r1: 7.2.4.1.5].
                # We'll strobe our less-severe "bad packet" indicator, but still reject the header.
                crc5_failed  = (expected_crc5 != packet.crc5)
                crc16_failed = (crc16.crc     != packet.crc16)
                with m.If(crc5_failed | crc16_failed):
                    m.d.comb += self.bad_packet.eq(1)

                # Our worst-case scenario is we're receiving a packet with an unexpected sequence
                # number; this indicates that we've lost sequence, and our device should move back to
                # into Recovery [USB3.2r1: 7.2.4.1.5].
                with m.Elif(packet.sequence_number != self.expected_sequence):
                    m.d.comb += self.bad_sequence.eq(1)

                # If neither of the above checks failed, we now know we have a valid header packet!
                # We'll output our packet, and then return to IDLE.
                with m.Else():
                    m.d.ss += [
                        self.new_packet  .eq(1),
                        self.packet      .eq(packet)
                    ]

                m.next = "WAIT_FOR_HPSTART"


        return m


class HeaderPacketReceiver(Elaboratable):
    """ Receiver-side Header Packet logic.

    This module handles all header-packet-reception related logic for the link layer; including
    header packet reception, buffering, flow control (credit management), and link command transmission.

    Attributes
    ----------
    sink: USBRawSuperSpeedStream(), input stream [monitor only]
        Stream that carries data from the physical layer, to be monitored.
    source: USBRawSuperSpeedStream(), output stream
        Stream that carries link commands from this unit down down to the physical layer.

    enable: Signal(), input
        When asserted, this unit will be enabled; and will be allowed to start
        transmitting link commands. Asserting this signal after a reset will perform
        a header sequence and link credit advertisement.
    usb_reset: Signal(), input
        Strobe; can be asserted to indicate that a USB reset has occurred, and sequencing
        should be restarted.

    queue: HeaderQueue(), output stream
        Stream carrying any header packets to be transmitted.

    retry_received: Signal(), input
        Strobe; should be asserted when the transmitter has seen a RETRY handshake.
    retry_required: Signal(), output
        Strobe; pulsed to indicate that we should send a RETRY handshake.

    link_command_sent: Signal(), output
        Strobe; pulses each time a link command is completed.
    keepalive_required: Signal(), input
        Strobe; when asserted; a keepalive packet will be generated.
    packet_received: Signal(), output
        Strobe; pulsed when an event occurs that should reset the USB3 "packet received" timers.
        This does *not* indicate valid data is present on the output :attr:``queue``; this has its
        own valid signal.
    bad_packet_received: Signal(), output
        Strobe; pulsed when a receive error occurs. For error counting at the link level.

    accept_power_state: Signal(), input
        Strobe; when pulsed, a LAU (Link-state acceptance) will be generated.
    reject_power_state: Signal(), input
        Strobe; when pulsed, a LXU (Link-state rejection) will be generated.
    acknowledge_power_state: Signal(), input
        Strobe; when pulsed, a LPMA (Link-state acknowledgement) will be generated.
    """

    SEQUENCE_NUMBER_WIDTH = 3

    def __init__(self, *, buffer_count=4, downstream_facing=False):
        self._buffer_count = buffer_count
        self._is_downstream_facing = downstream_facing

        #
        # I/O port
        #
        self.sink                    = USBRawSuperSpeedStream()
        self.source                  = USBRawSuperSpeedStream()

        # Simple controls.
        self.enable                  = Signal()
        self.usb_reset               = Signal()

        # Header Packet Queue
        self.queue                   = HeaderQueue()

        # Event signaling.
        self.retry_received          = Signal()
        self.lrty_pending            = Signal()
        self.retry_required          = Signal()
        self.recovery_required       = Signal()

        self.link_command_sent       = Signal()
        self.keepalive_required      = Signal()
        self.packet_received         = Signal()
        self.bad_packet_received     = Signal()

        self.accept_power_state      = Signal()
        self.reject_power_state      = Signal()
        self.acknowledge_power_state = Signal()


    def elaborate(self, platform):
        m = Module()

        #
        # Sequence tracking.
        #

        # Keep track of which sequence number we expect to see.
        expected_sequence_number = Signal(self.SEQUENCE_NUMBER_WIDTH)

        # Keep track of which credit we'll need to issue next...
        next_credit_to_issue     = Signal(range(self._buffer_count))

        # ... and which header we'll need to ACK next.
        # We'll start with the maximum number, so our first advertisement wraps us back around to zero.
        next_header_to_ack       = Signal.like(expected_sequence_number, init=-1)


        #
        # Task "queues".
        #

        # Keep track of how many header received acknowledgements (LGOODs) we need to send.
        acks_to_send      = Signal(range(self._buffer_count + 1), init=1)
        enqueue_ack       = Signal()
        dequeue_ack       = Signal()

        with m.If(enqueue_ack & ~dequeue_ack):
            m.d.ss += acks_to_send.eq(acks_to_send + 1)
        with m.If(dequeue_ack & ~enqueue_ack):
            m.d.ss += acks_to_send.eq(acks_to_send - 1)


        # Keep track of how many link credits we've yet to free.
        # We'll start with every one of our buffers marked as "pending free"; this ensures
        # we perform our credit restoration properly.
        credits_to_issue  = Signal.like(acks_to_send, init=self._buffer_count)
        enqueue_credit_issue = Signal()
        dequeue_credit_issue = Signal()

        with m.If(enqueue_credit_issue & ~dequeue_credit_issue):
            m.d.ss += credits_to_issue.eq(credits_to_issue + 1)
        with m.If(dequeue_credit_issue & ~enqueue_credit_issue):
            m.d.ss += credits_to_issue.eq(credits_to_issue - 1)

        # Keep track of whether we should be sending an LBAD.
        lbad_pending = Signal()

        # Keep track of whether a retry has been requested.
        lrty_pending = self.lrty_pending
        with m.If(self.retry_required):
            m.d.ss += lrty_pending.eq(1)

        # Keep track of whether a keepalive has been requested.
        keepalive_pending = Signal()
        with m.If(self.keepalive_required):
            m.d.ss += keepalive_pending.eq(1)

        # Keep track of whether we're expected to send an power state response.
        lau_pending  = Signal()
        lxu_pending  = Signal()
        lpma_pending = Signal()

        with m.If(self.accept_power_state):
            m.d.ss += lau_pending.eq(1)
        with m.If(self.reject_power_state):
            m.d.ss += lxu_pending.eq(1)
        with m.If(self.acknowledge_power_state):
            m.d.ss += lpma_pending.eq(1)

        # Keep track of the last value of the enable signal
        last_enable = Signal()
        m.d.ss     += last_enable.eq(self.enable)

        #
        # Header Packet Buffers
        #

        # Track which buffer will be filled next.
        read_pointer      = Signal(range(self._buffer_count))
        write_pointer     = Signal.like(read_pointer)

        # Track how many buffers we currently have in use.
        buffers_filled    = Signal.like(credits_to_issue, init=0)
        reserve_buffer    = Signal()
        release_buffer    = Signal()

        with m.If(reserve_buffer & ~release_buffer):
            m.d.ss += buffers_filled.eq(buffers_filled + 1)
        with m.If(release_buffer & ~reserve_buffer):
            m.d.ss += buffers_filled.eq(buffers_filled - 1)

        # Create buffers to receive any incoming header packets.
        buffers = Array(HeaderPacket() for _ in range(self._buffer_count))


        #
        # Packet reception (physical layer -> link layer).
        #

        # Flag that determines when we should ignore packets.
        #
        # After a receive error, we'll want to ignore all packets until we see a "retry"
        # link command; so we don't receive packets out of order.
        ignore_packets = Signal()

        # Create our raw packet parser / receiver.
        m.submodules.receiver = rx = RawHeaderPacketReceiver()
        m.d.comb += [
            # Our receiver passively monitors the data received for header packets.
            rx.sink                   .tap(self.sink),

            # Ensure it's always up to date about what sequence numbers we expect.
            rx.expected_sequence      .eq(expected_sequence_number),

            # If we ever get a bad header packet sequence, we're required to retrain
            # the link [USB3.2r1: 7.2.4.1.5]. Pass the event through directly.
            self.recovery_required    .eq(rx.bad_sequence & ~ignore_packets),

            # Notify the link layer when packets are received, for keeping track of our timers.
            self.packet_received      .eq(rx.new_packet),

            # Notify the link layer if any bad packets are received; for diagnostics.
            self.bad_packet_received  .eq(rx.bad_packet)
        ]


        # If we receive a valid packet, it's time for us to buffer it!
        with m.If(rx.new_packet & ~ignore_packets):
            m.d.ss += [
                # Load our header packet into the next write buffer...
                buffers[write_pointer]    .eq(rx.packet),

                # ... advance to the next buffer and sequence number...
                write_pointer             .eq(write_pointer + 1),
                expected_sequence_number  .eq(expected_sequence_number + 1),
            ]
            m.d.comb += [
                # ... mark the buffer space as occupied by valid data ...
                reserve_buffer            .eq(1),

                # ... and queue an ACK for this packet.
                enqueue_ack               .eq(1)
            ]


        # If we receive a bad packet, we'll need to request that the other side re-send.
        # The rules for this are summarized in [USB3.2r1: 7.2.4.1.5], and in comments below.
        with m.If(rx.bad_packet & ~ignore_packets):


            m.d.ss += [
                # First, we'll need to schedule transmission of an LBAD, which notifies the other
                # side that we received a bad packet; and that it'll need to transmit all unack'd
                # header packets to us again.
                lbad_pending    .eq(1),

                # Next, we'll need to make sure we don't receive packets out of sequence. This means
                # we'll have to start ignoring packets until the other side responds to the LBAD.
                # The other side respond with an Retry link command (LRTY) once it's safe for us to
                # pay attention to packets again.
                ignore_packets  .eq(1)
            ]


        # Finally, if we receive a Retry link command, this means we no longer need to ignore packets.
        # This typically happens in response to us sending an LBAD and marking future packets as ignored.
        with m.If(self.retry_received):
            m.d.ss += ignore_packets.eq(0)


        #
        # Packet delivery (link layer -> physical layer).
        #
        m.d.comb += [
            # As long as we have at least one buffer filled, we have header packets pending.
            self.queue.valid    .eq(buffers_filled > 0),

            # Always provide the value of our oldest packet out to our consumer.
            self.queue.header    .eq(buffers[read_pointer])
        ]


        # If the protocol layer is marking one of our packets as consumed, we no longer
        # need to buffer it -- it's the protocol layer's problem, now!
        with m.If(self.queue.valid & self.queue.ready):

            # Move on to reading from the next buffer in sequence.
            m.d.ss += read_pointer.eq(read_pointer + 1)

            m.d.comb += [
                # First, we'll free the buffer associated with the relevant packet...
                release_buffer        .eq(1),

                # ... and request that our link partner be notified of the new space.
                enqueue_credit_issue  .eq(1)
            ]


        #
        # Link command generation.
        #
        m.submodules.lc_generator = lc_generator = LinkCommandGenerator()
        m.d.comb += [
            self.source             .stream_eq(lc_generator.source),
            self.link_command_sent  .eq(lc_generator.done),
        ]


        with m.FSM(domain="ss"):

            # DISPATCH_COMMAND -- the state in which we identify any pending link commands necessary,
            # and then move to the state in which we'll send them.
            with m.State("DISPATCH_COMMAND"):

                with m.If(self.enable):
                    # NOTE: the order below is important; changing it can easily break things:
                    # - ACKS must come before credits, as we must send an LGOOD before we send our initial credits.
                    # - LBAD must come after ACKs and credit management, as all scheduled ACKs need to be
                    #   sent to the other side for the LBAD to have the correct semantic meaning.

                    with m.If(lrty_pending):
                        m.next = "SEND_LRTY"

                    # If we have acknowledgements to send, send them.
                    with m.Elif(acks_to_send):
                        m.next = "SEND_ACKS"

                    # If we have link credits to issue, move to issuing them to the other side.
                    with m.Elif(credits_to_issue):
                        m.next = "ISSUE_CREDITS"

                    # If we need to send an LBAD, do so.
                    with m.Elif(lbad_pending):
                        m.next = "SEND_LBAD"

                    # If we need to send a link power-state command, do so.
                    with m.Elif(lxu_pending):
                        m.next = "SEND_LXU"

                    # If we need to send a keepalive, do so.
                    with m.Elif(keepalive_pending):
                        m.next = "SEND_KEEPALIVE"



                # Once we've become disabled, we'll want to prepare for our next enable.
                # This means preparing for our advertisement, by:
                with m.If((last_enable & ~self.enable) | self.usb_reset):
                    m.d.ss += [
                        # -Resetting our pending ACKs to 1, so we perform an sequence number advertisement
                        #  when we're next enabled.
                        acks_to_send          .eq(1),

                        # -Decreasing our next sequence number; so we maintain a continuity of sequence numbers
                        #  without counting the advertising one. This doesn't seem to be be strictly necessary
                        #  per the spec; but seem to make analyzers happier, so we'll go with it.
                        next_header_to_ack    .eq(next_header_to_ack - 1),

                        # - Clearing all of our buffers.
                        read_pointer          .eq(0),
                        write_pointer         .eq(0),
                        buffers_filled        .eq(0),

                        # - Preparing to re-issue all of our buffer credits.
                        next_credit_to_issue  .eq(0),
                        credits_to_issue      .eq(self._buffer_count),

                        # - Clear our pending events.
                        lrty_pending          .eq(0),
                        lbad_pending          .eq(0),
                        keepalive_pending     .eq(0),
                        ignore_packets        .eq(0)
                    ]

                    # If this is a USB Reset, also reset our sequences.
                    with m.If(self.usb_reset):
                        m.d.ss += [
                            expected_sequence_number  .eq(0),
                            next_header_to_ack        .eq(-1)
                        ]


            # SEND_ACKS -- a valid header packet has been received, or we're advertising
            # our initial sequence number; send an LGOOD packet.
            with m.State("SEND_ACKS"):

                # Send an LGOOD command, acknowledging the last received packet header.
                m.d.comb += [
                    lc_generator.generate      .eq(1),
                    lc_generator.command       .eq(LinkCommand.LGOOD),
                    lc_generator.subtype       .eq(next_header_to_ack)
                ]

                # Wait until our link command is done, and then move on.
                with m.If(lc_generator.done):
                    # Move to the next header packet in the sequence, and decrease
                    # the number of outstanding ACKs.
                    m.d.comb += dequeue_ack         .eq(1)
                    m.d.ss   += next_header_to_ack  .eq(next_header_to_ack + 1)

                    # If this was the last ACK we had to send, move back to our dispatch state.
                    with m.If(acks_to_send == 1):
                        m.next = "DISPATCH_COMMAND"


            # ISSUE_CREDITS -- header packet buffers have been freed; and we now need to notify the
            # other side, so it knows we have buffers available.
            with m.State("ISSUE_CREDITS"):

                # Send an LCRD command, indicating that we have a free buffer.
                m.d.comb += [
                    lc_generator.generate      .eq(1),
                    lc_generator.command       .eq(LinkCommand.LCRD),
                    lc_generator.subtype       .eq(next_credit_to_issue)
                ]

                # Wait until our link command is done, and then move on.
                with m.If(lc_generator.done):
                    # Move to the next credit...
                    m.d.comb += dequeue_credit_issue  .eq(1)
                    m.d.ss   += next_credit_to_issue  .eq(next_credit_to_issue + 1)

                    # If this was the last credit we had to issue, move back to our dispatch state.
                    with m.If(credits_to_issue == 1):
                        m.next = "DISPATCH_COMMAND"


            # SEND_LBAD -- we've received a bad header packet; we'll need to let the other side know.
            with m.State("SEND_LBAD"):
                m.d.comb += [
                    lc_generator.generate      .eq(1),
                    lc_generator.command       .eq(LinkCommand.LBAD),
                ]

                # Once we've sent the LBAD, we can mark is as no longer pending and return to our dispatch.
                # (We can't ever have multiple LBADs queued up; as we ignore future packets after sending one.)
                with m.If(lc_generator.done):
                    m.d.ss += lbad_pending.eq(0)
                    m.next = "DISPATCH_COMMAND"


            # SEND_LRTY -- our transmitter has requested that we send an retry indication to the other side.
            # We'll do our transmitter a favor and do so.
            with m.State("SEND_LRTY"):
                m.d.comb += [
                    lc_generator.generate      .eq(1),
                    lc_generator.command       .eq(LinkCommand.LRTY)
                ]

                with m.If(lc_generator.done):
                    m.d.ss += lrty_pending.eq(0)
                    m.next = "DISPATCH_COMMAND"



            # SEND_KEEPALIVE -- our link layer timer has requested that we send a keep-alive,
            # indicating that we're still in U0 and the link is still good. Do so.
            with m.State("SEND_KEEPALIVE"):

                # Send the correct packet type for the direction our port is facing.
                command = LinkCommand.LDN if self._is_downstream_facing else LinkCommand.LUP

                m.d.comb += [
                    lc_generator.generate      .eq(1),
                    lc_generator.command       .eq(command)
                ]

                # Once we've send the keepalive, we can mark is as no longer pending and return to our dispatch.
                # (There's no sense in sending repeated keepalives; one gets the message across.)
                with m.If(lc_generator.done):
                    m.d.ss += keepalive_pending.eq(0)
                    m.next = "DISPATCH_COMMAND"


            # SEND_LXU -- we're being instructed to reject a requested power-state transfer.
            # We'll send an LXU packet to inform the other side of the rejection.
            with m.State("SEND_LXU"):
                m.d.comb += [
                    lc_generator.generate      .eq(1),
                    lc_generator.command       .eq(LinkCommand.LXU)
                ]

                with m.If(lc_generator.done):
                    m.d.ss += lxu_pending.eq(0)
                    m.next = "DISPATCH_COMMAND"

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" UART interface gateware."""

from amaranth   import Elaboratable, Module, Signal, Cat

from ..stream       import StreamInterface


class UARTTransmitter(Elaboratable):
    """ Simple UART transitter.

    Intended for communicating with the debug controller; currently assumes 8n1.

    Attributes
    ----------

    tx: Signal(), output
        The UART output.
    driving: Signal(), output
        True iff the UART is in the middle of driving data. In some cases, it's desireable
        to have the UART drive the line only when it is actively sending; letting a pull
        resistor handle pulling the line to idle. This line can be used to determine when the
        line should be driven.

    stream: input stream
        The stream carrying the data to be sent.

    idle: Signal(), output
        Asserted when the transmitter is idle; and thus pulsing `send_active`
        will start a new transmission.

    Parameters
    ------------
    divisor: int
        The number of `sync` clock cycles per bit period.
    """

    START_BIT = 0
    STOP_BIT  = 1

    def __init__(self, *, divisor):
        self.divisor = divisor

        #
        # I/O port
        #
        self.tx              = Signal(init=1)
        self.driving         = Signal()
        self.stream          = StreamInterface()

        self.idle            = Signal()


    def elaborate(self, platform):
        m = Module()

        # Baud generator.
        baud_counter = Signal(range(0, self.divisor))

        # Tx shift register; holds our data, a start, and a stop bit.
        bits_per_frame = len(self.stream.payload) + 2
        data_shift     = Signal(bits_per_frame)
        bits_to_send   = Signal(range(0, len(data_shift)))

        # Create an internal signal equal to our input data framed with a start/stop bit.
        framed_data_in = Cat(self.START_BIT, self.stream.payload, self.STOP_BIT)


        with m.FSM() as f:
            m.d.comb += self.idle.eq(f.ongoing('IDLE'))

            # IDLE: transmitter is waiting for input
            with m.State("IDLE"):
                m.d.comb += [
                    self.tx            .eq(1),
                    self.stream.ready  .eq(1)
                ]


                # Once we get a send request, fill in our shift register, and start shifting.
                with m.If(self.stream.valid):
                    m.d.sync += [
                        baud_counter      .eq(self.divisor   - 1),
                        bits_to_send      .eq(len(data_shift) - 1),
                        data_shift        .eq(framed_data_in),
                    ]

                    m.next = "TRANSMIT"


            # TRANSMIT: actively shift out start/data/stop
            with m.State("TRANSMIT"):
                m.d.sync += baud_counter  .eq(baud_counter - 1)
                m.d.comb += [
                    self.tx       .eq(data_shift[0]),
                    self.driving  .eq(1)
                ]

                # If we've finished a bit period...
                with m.If(baud_counter == 0):
                    m.d.sync += baud_counter.eq(self.divisor - 1)

                    # ... if we have bits left to send, move to the next one.
                    with m.If(bits_to_send > 0):
                        m.d.sync += [
                            bits_to_send .eq(bits_to_send - 1),
                            data_shift   .eq(data_shift[1:])
                        ]

                    # Otherwise, complete the frame.
                    with m.Else():
                        m.d.comb += self.stream.ready.eq(1)

                        # If we still have data to send, move to the next byte...
                        with m.If(self.stream.valid):
                            m.d.sync += [
                                bits_to_send  .eq(bits_per_frame - 1),
                                data_shift    .eq(framed_data_in),
                            ]

                        # ... otherwise, move to our idle state.
                        with m.Else():
                            m.next = "IDLE"


        return m



class UARTMultibyteTransmitter(Elaboratable):
    """ UART transmitter capable of sending wide words.

    Intended for communicating with the debug controller; currently assumes 8n1.
    Transmits our words little-endian.

    Attributes
    ----------

    tx: Signal(), output
        The UART output.
    stream: input stream
        The data to be transmitted.

    accepted: Signal(), output
        Strobe that indicates when the `data` word has been latched in;
        and the next data byte can be presented.
    idle: Signal(), output
        Asserted when the transmitter is idle; and thus pulsing `send_active`
        will start a new transmission.

    Parameters
    ------------
    byte_width: int
        The number of bytes to be accepted at once.

    divisor: int
        The number of `sync` clock cycles per bit period.
    """
    def __init__(self, *, byte_width, divisor):
        self.byte_width = byte_width
        self.divisor = divisor

        #
        # I/O port
        #
        self.tx              = Signal(init=1)
        self.stream          = StreamInterface(payload_width=byte_width * 8)

        self.idle            = Signal()


    def elaborate(self, platform):
        m = Module()

        # Create our core UART transmitter.
        m.submodules.uart = uart = UARTTransmitter(divisor=self.divisor)

        # We'll put each word to be sent through an shift register
        # that shifts out words a byte at a time.
        data_shift = Signal.like(self.stream.payload)

        # Count how many bytes we have left to send.
        bytes_to_send = Signal(range(0, self.byte_width + 1))

        m.d.comb += [

            # Connect our transmit output directly through.
            self.tx.eq(uart.tx),

            # Always provide our UART with the least byte of our shift register.
            uart.stream.payload.eq(data_shift[0:8])
        ]



        with m.FSM() as f:
            m.d.comb += self.idle.eq(f.ongoing('IDLE'))

            # IDLE: transmitter is waiting for input
            with m.State("IDLE"):
                m.d.comb += self.stream.ready.eq(1)

                # Once we get a send request, fill in our shift register, and start shifting.
                with m.If(self.stream.valid):
                    m.d.sync += [
                        data_shift         .eq(self.stream.payload),
                        bytes_to_send      .eq(self.byte_width - 1),
                    ]
                    m.next = "TRANSMIT"


            # TRANSMIT: actively send each of the bytes of our word
            with m.State("TRANSMIT"):
                m.d.comb += uart.stream.valid.eq(1)

                # Once the UART is accepting our input...
                with m.If(uart.stream.ready):

                    # ... if we have bytes left to send, move to the next one.
                    with m.If(bytes_to_send > 0):
                        m.d.sync += [
                            bytes_to_send .eq(bytes_to_send - 1),
                            data_shift    .eq(data_shift[8:]),
                        ]

                    # Otherwise, complete the frame.
                    with m.Else():
                        m.d.comb += self.stream.ready.eq(1)

                        # If we still have data to send, move to the next byte...
                        with m.If(self.stream.valid):
                            m.d.sync += [
                                bytes_to_send      .eq(self.byte_width - 1),
                                data_shift         .eq(self.stream.payload),
                            ]

                        # ... otherwise, move to our idle state.
                        with m.Else():
                            m.next = "IDLE"


        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
"""
Contains the organizing hardware used to add USB3 Device functionality
to your own designs; including the core :class:`USBSuperSpeedDevice` class.
"""

import logging

from amaranth import *

from usb_protocol.emitters import DeviceDescriptorCollection

# USB3 Protocol Stack
from .physical             import USB3PhysicalLayer
from .link                 import USB3LinkLayer
from .protocol             import USB3ProtocolLayer
from .endpoints            import USB3ControlEndpoint
from .protocol.endpoint    import SuperSpeedEndpointMultiplexer

# Temporary
from ..stream              import USBRawSuperSpeedStream, SuperSpeedStreamInterface


class USBSuperSpeedDevice(Elaboratable):
    """ Core gateware common to all LUNA USB3 devices. """

    def __init__(self, *, phy, sync_frequency=None):
        self._phy = phy
        self._sync_frequency = sync_frequency

        # Create a collection of endpoints for this device.
        self._endpoints = []

        #
        # I/O port
        #

        # General status signals.
        self.link_trained   = Signal()
        self.link_in_reset  = Signal()

        # Temporary, debug signals.
        self.rx_data_tap         = USBRawSuperSpeedStream()
        self.tx_data_tap         = USBRawSuperSpeedStream()

        self.ep_tx_stream        = SuperSpeedStreamInterface()
        self.ep_tx_length        = Signal(range(1024 + 1))


    def add_endpoint(self, endpoint):
        """ Adds an endpoint interface to the device.

        Parameters
        ----------
        endpoint: Elaborateable
            The endpoint interface to be added. Can be any piece of gateware with a
            :class:`EndpointInterface` attribute called ``interface``.
        """
        self._endpoints.append(endpoint)


    def add_standard_control_endpoint(self, descriptors: DeviceDescriptorCollection):
        """ Adds a control endpoint with standard request handlers to the device.

        Parameters
        ----------
        descriptors: DeviceDescriptorCollection
            The descriptors to use for this device.

        Return value
        ------------
        The endpoint object created.
        """

        # TODO: split out our standard request handlers

        control_endpoint = USB3ControlEndpoint()
        control_endpoint.add_standard_request_handlers(descriptors)
        self.add_endpoint(control_endpoint)

        return control_endpoint



    def elaborate(self, platform):
        m = Module()

        # Figure out the frequency of our ``sync`` domain, for e.g. PHY bringup timing.
        # We'll default to the platform's default frequency if none was provided.
        sync_frequency = self._sync_frequency
        if sync_frequency is None:
            sync_frequency = platform.default_clk_frequency

        #
        # Global device state.
        #

        # Stores the device's current address. Used to identify which packets are for us.
        address       = Signal(7, init=0)

        # Stores the device's current configuration. Defaults to unconfigured.
        configuration = Signal(8, init=0)


        #
        # Physical layer.
        #
        m.submodules.physical = physical = USB3PhysicalLayer(
            phy            = self._phy,
            sync_frequency = sync_frequency
        )

        #
        # Link layer.
        #
        m.submodules.link = link = USB3LinkLayer(physical_layer=physical)
        m.d.comb += [
            self.link_trained     .eq(link.trained),
            self.link_in_reset    .eq(link.in_reset),

            link.current_address  .eq(address)
        ]

        #
        # Protocol layer.
        #
        m.submodules.protocol = protocol = USB3ProtocolLayer(link_layer=link)
        m.d.comb += [
            protocol.current_address        .eq(address),
            protocol.current_configuration  .eq(configuration)
        ]


        #
        # Application layer.
        #

        # Create our endpoint multiplexer...
        m.submodules.endpoint_mux = endpoint_mux = SuperSpeedEndpointMultiplexer()
        endpoint_collection = endpoint_mux.shared

        m.d.comb += [
            # Receive interface.
            endpoint_collection.rx                          .tap(protocol.endpoint_interface.rx),
            endpoint_collection.rx_header                   .eq(protocol.endpoint_interface.rx_header),
            endpoint_collection.rx_complete                 .eq(protocol.endpoint_interface.rx_complete),
            endpoint_collection.rx_invalid                  .eq(protocol.endpoint_interface.rx_invalid),

            # Transmit interface.
            protocol.endpoint_interface.tx                  .stream_eq(endpoint_collection.tx),
            protocol.endpoint_interface.tx_zlp              .eq(endpoint_collection.tx_zlp),
            protocol.endpoint_interface.tx_length           .eq(endpoint_collection.tx_length),
            protocol.endpoint_interface.tx_endpoint_number  .eq(endpoint_collection.tx_endpoint_number),
            protocol.endpoint_interface.tx_sequence_number  .eq(endpoint_collection.tx_sequence_number),
            protocol.endpoint_interface.tx_direction        .eq(endpoint_collection.tx_direction),

            # Handshake interface.
            protocol.endpoint_interface.handshakes_out      .connect(endpoint_collection.handshakes_out),
            protocol.endpoint_interface.handshakes_in       .connect(endpoint_collection.handshakes_in)
        ]


        # If an endpoint wants to update our address or configuration, accept the update.
        with m.If(endpoint_collection.address_changed):
            m.d.ss += address.eq(endpoint_collection.new_address)
        with m.If(endpoint_collection.config_changed):
            m.d.ss += configuration.eq(endpoint_collection.new_config)


        # Finally, add each of our endpoints to this module and our multiplexer.
        for endpoint in self._endpoints:

            # Create a display name for the endpoint...
            name = endpoint.__class__.__name__
            if hasattr(m.submodules, name):
                name = f"{name}_{id(endpoint)}"

            # ... and add it, both as a submodule and to our multiplexer.
            endpoint_mux.add_interface(endpoint.interface)
            m.submodules[name] = endpoint


        #
        # Reset handling.
        #

        # Restore ourselves to our unconfigured state when a reset occurs.
        with m.If(link.in_reset):
            m.d.ss += [
                address        .eq(0),
                configuration  .eq(0)
            ]


        #
        # Debug helpers.
        #

        # Tap our transmit and receive lines, so they can be externally analyzed.
        m.d.comb += [
            self.rx_data_tap   .tap(physical.source),
            self.tx_data_tap   .tap(physical.sink),

            self.ep_tx_stream  .tap(protocol.endpoint_interface.tx, tap_ready=True),
            self.ep_tx_length  .eq(protocol.endpoint_interface.tx_length)
        ]


        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
"""
Contains the organizing hardware used to add USB Device functionality
to your own designs; including the core :class:`USBDevice` class.
"""

import logging

from luna                      import configure_default_logging

from amaranth                  import Signal, Module, Elaboratable, Const
from usb_protocol.emitters     import DeviceDescriptorCollection

from ...interface.ulpi         import UTMITranslator
from ...interface.utmi         import UTMIInterfaceMultiplexer
from ...interface.gateware_phy import GatewarePHY

from .packet                   import USBTokenDetector, USBHandshakeGenerator, USBDataPacketCRC
from .packet                   import USBInterpacketTimer, USBDataPacketGenerator, USBHandshakeDetector
from .packet                   import USBDataPacketReceiver
from .reset                    import USBResetSequencer

from .endpoint                 import USBEndpointMultiplexer
from .control                  import USBControlEndpoint


class USBDevice(Elaboratable):
    """ Core gateware common to all LUNA USB2 devices.

    The ``USBDevice`` module contains the low-level communications hardware necessary to implement a USB device;
    including hardware for maintaining device state, detecting events, reading data from the host, and generating
    responses.

    This class can be instantiated directly, and used to build a USB device,
    or can be subclassed to create custom device types.

    To configure a ``USBDevice`` from a CPU or other wishbone master, see :class:`USBDeviceController`;
    which can easily be attached using its `attach` method.


    Parameters
    ----------

    bus: [UTMI interface, ULPI Interface]
        The UTMI or ULPI PHY connection to be used for communications.

    handle_clocking: bool, Optional
        True iff we should attempt to connect up the `usb` clock domain to the PHY
        automatically based on the clk signals's I/O direction. This option may not work
        for non-simple connections; in which case you will need to connect the clock signal
        yourself.


    Attributes
    ----------

    connect: Signal(), input
        Held high to keep the current USB device connected; or held low to disconnect.
    low_speed_only: Signal(), input
        If high, the device will operate at low speed.
    full_speed_only: Signal(), input
        If high, the device will be prohibited from operating at high speed.

    frame_number: Signal(11), output
        The current USB frame number.
    microframe_number: Signal(3), output
        The current USB microframe number. Always 0 on non-HS connections.
    sof_detected: Signal(), output
        Pulses for one cycle each time a SOF is detected; and thus our frame number has changed.
    new_frame: Signal(), output
        Strobe that indicates a new frame (not microframe) is detected.

    reset_detected: Signal(), output
        Asserted when the USB device receives a bus reset.

    # State signals.
    suspended: Signal(), output
        High when the device is in USB suspend. This can be (and by the spec must be) used to trigger
        the device to enter lower-power states.

    tx_activity_led: Signal(), output
        Signal that can be used to drive an activity LED for TX.
    rx_activity_led: Signal(), output
        Signal that can be used to drive an activity LED for RX.

    """

    def __init__(self, *, bus, handle_clocking=True):
        """
        Parameters:
        """

        # If this looks more like a ULPI bus than a UTMI bus, translate it.
        if hasattr(bus, 'dir'):
            self.utmi       = UTMITranslator(ulpi=bus, handle_clocking=handle_clocking)
            self.bus_busy   = self.utmi.busy
            self.translator = self.utmi
            self.always_fs  = False
            self.data_clock = 60e6

        # If this looks more like raw I/O connections than a UTMI bus, create a pure-gatware
        # PHY to drive the raw I/O signals.
        elif hasattr(bus, 'd_n'):
            self.utmi       = GatewarePHY(io=bus)
            self.bus_busy   = Const(0)
            self.translator = self.utmi
            self.always_fs  = True
            self.data_clock = 12e6

        # Otherwise, use it directly.
        # Note that since a true UTMI interface has separate Tx/Rx/control
        # interfaces, we don't need to care about bus 'busyness'; so we'll
        # set it to a const zero.
        else:
            self.utmi       = bus
            self.bus_busy   = Const(0)
            self.translator = None
            self.always_fs  = True
            self.data_clock = 12e6

        #
        # I/O port
        #
        self.connect           = Signal()
        self.low_speed_only    = Signal()
        self.full_speed_only   = Signal()

        self.frame_number      = Signal(11)
        self.microframe_number = Signal(3)
        self.sof_detected      = Signal()
        self.new_frame         = Signal()
        self.reset_detected    = Signal()

        self.speed             = Signal(2)
        self.suspended         = Signal()
        self.tx_activity_led   = Signal()
        self.rx_activity_led   = Signal()

        #
        # Internals.
        #
        self._endpoints = []

        # Try to retrieve the bus name, needed for USB device hooks from platform
        self._bus_name = None
        try:
            if hasattr(bus, 'name'):
                self._bus_name = bus.name
            elif hasattr(bus, 'clk'):
                # PureInterface does not have a name attribute, but we can use the
                # first item of a path tuple
                self._bus_name = bus.clk.path[0]
        except (AttributeError, TypeError):
            pass


    def add_endpoint(self, endpoint):
        """ Adds an endpoint interface to the device.

        Parameters
        ----------
        endpoint: Elaborateable
            The endpoint interface to be added. Can be any piece of gateware with a
            :class:`EndpointInterface` attribute called ``interface``.
        """
        self._endpoints.append(endpoint)


    def add_control_endpoint(self):
        """ Adds a basic control endpoint to the device.

        Does not add any request handlers. If you want standard request handlers;
        :attr:`add_standard_control_endpoint` automatically adds standard request handlers.

        Returns
        -------
        Returns the endpoint object for the control endpoint.
        """
        control_endpoint = USBControlEndpoint(utmi=self.utmi)
        self.add_endpoint(control_endpoint)

        return control_endpoint


    def add_standard_control_endpoint(self, descriptors: DeviceDescriptorCollection, **kwargs):
        """ Adds a control endpoint with standard request handlers to the device.

        Parameters will be passed on to StandardRequestHandler.

        Return value
        ------------
        The endpoint object created.
        """

        # Create our endpoint, and add standard descriptors to it.
        control_endpoint = USBControlEndpoint(utmi=self.utmi)
        control_endpoint.add_standard_request_handlers(descriptors, **kwargs)
        self.add_endpoint(control_endpoint)

        return control_endpoint



    def elaborate(self, platform):
        m = Module()

        # Call any hooks registered for the bus in the platform definition.
        if hasattr(platform, "usb_device_hooks"):
            hook = platform.usb_device_hooks.get(self._bus_name)
            if hook is not None:
                hook(self, m)

        # If we have a bus translator, include it in our submodules.
        if self.translator:
            m.submodules.translator = self.translator


        #
        # Internal device state.
        #

        # Stores the device's current address. Used to identify which packets are for us.
        address       = Signal(7, init=0)

        # Stores the device's current configuration. Defaults to unconfigured.
        configuration = Signal(8, init=0)


        #
        # Internal interconnections.
        #

        # Create our reset sequencer, which will be in charge of detecting USB port resets,
        # detecting high-speed hosts, and communicating that we are a high speed device.
        m.submodules.reset_sequencer = reset_sequencer = USBResetSequencer()

        m.d.comb += [
            reset_sequencer.bus_busy        .eq(self.bus_busy),

            reset_sequencer.vbus_connected  .eq(~self.utmi.session_end),
            reset_sequencer.line_state      .eq(self.utmi.line_state),

            reset_sequencer.disconnect      .eq(~self.connect),
        ]


        # Create our internal packet components:
        # - A token detector, which will identify and parse the tokens that start transactions.
        # - A data transmitter, which will transmit provided data streams.
        # - A data receiver, which will receive data from UTMI and convert it into streams.
        # - A handshake generator, which will assist in generating response packets.
        # - A handshake detector, which detects handshakes generated by the host.
        # - A data CRC16 handler, which will compute data packet CRCs.
        # - An interpacket delay timer, which will enforce interpacket delays.
        m.submodules.token_detector      = token_detector      = \
            USBTokenDetector(utmi=self.utmi, domain_clock=self.data_clock, fs_only=self.always_fs)
        m.submodules.transmitter         = transmitter         = USBDataPacketGenerator()
        m.submodules.receiver            = receiver            = USBDataPacketReceiver(utmi=self.utmi)
        m.submodules.handshake_generator = handshake_generator = USBHandshakeGenerator()
        m.submodules.handshake_detector  = handshake_detector  = USBHandshakeDetector(utmi=self.utmi)
        m.submodules.data_crc            = data_crc            = USBDataPacketCRC()
        m.submodules.timer               = timer               = \
            USBInterpacketTimer(domain_clock=self.data_clock, fs_only=self.always_fs)

        # Connect our transmitter/receiver to our CRC generator.
        data_crc.add_interface(transmitter.crc)
        data_crc.add_interface(receiver.data_crc)

        # Connect our receiver to our timer.
        timer.add_interface(receiver.timer)

        m.d.comb += [
            # Ensure our token detector only responds to tokens addressed to us.
            token_detector.address  .eq(address),

            # Hook up our data_crc to our receive inputs.
            data_crc.rx_data        .eq(self.utmi.rx_data),
            data_crc.rx_valid       .eq(self.utmi.rx_valid),

            # Connect our state signals to our subordinate components.
            token_detector.speed    .eq(self.speed),
            timer.speed             .eq(self.speed)
        ]

        #
        # Endpoint connections.
        #

        # Create our endpoint multiplexer...
        m.submodules.endpoint_mux = endpoint_mux = USBEndpointMultiplexer()
        endpoint_collection = endpoint_mux.shared

        # Connect our timer and CRC interfaces.
        timer.add_interface(endpoint_collection.timer)
        data_crc.add_interface(endpoint_collection.data_crc)

        m.d.comb += [
            # Low-level hardware interface.
            token_detector.interface                   .connect(endpoint_collection.tokenizer),
            handshake_detector.detected                .connect(endpoint_collection.handshakes_in),

            # Device state.
            endpoint_collection.speed                  .eq(self.speed),
            endpoint_collection.active_config          .eq(configuration),
            endpoint_collection.active_address         .eq(address),

            # Receive interface.
            receiver.stream                            .connect(endpoint_collection.rx),
            endpoint_collection.rx_complete            .eq(receiver.packet_complete),
            endpoint_collection.rx_invalid             .eq(receiver.crc_mismatch),
            endpoint_collection.rx_ready_for_response  .eq(receiver.ready_for_response),
            endpoint_collection.rx_pid_toggle          .eq(receiver.active_pid[3]),

            # Transmit interface.
            endpoint_collection.tx                     .attach(transmitter.stream),
            handshake_generator.issue_ack              .eq(endpoint_collection.handshakes_out.ack),
            handshake_generator.issue_nak              .eq(endpoint_collection.handshakes_out.nak),
            handshake_generator.issue_stall            .eq(endpoint_collection.handshakes_out.stall),
            transmitter.data_pid                       .eq(endpoint_collection.tx_pid_toggle),
        ]

        # If an endpoint wants to update our address or configuration, accept the update.
        with m.If(endpoint_collection.address_changed):
            m.d.usb += address.eq(endpoint_collection.new_address)
        with m.If(endpoint_collection.config_changed):
            m.d.usb += configuration.eq(endpoint_collection.new_config)


        # Finally, add each of our endpoints to this module and our multiplexer.
        for endpoint in self._endpoints:

            # Create a display name for the endpoint...
            name = endpoint.__class__.__name__
            if hasattr(m.submodules, name):
                name = f"{name}_{id(endpoint)}"

            # ... and add it, both as a submodule and to our multiplexer.
            endpoint_mux.add_interface(endpoint.interface)
            m.submodules[name] = endpoint


        #
        # Transmitter multiplexing.
        #

        # Create a multiplexer that will arbitrate access to the transmit lines.
        m.submodules.tx_multiplexer = tx_multiplexer = UTMIInterfaceMultiplexer()

        # Connect each of our transmitters.
        tx_multiplexer.add_input(reset_sequencer.tx)
        tx_multiplexer.add_input(transmitter.tx)
        tx_multiplexer.add_input(handshake_generator.tx)

        m.d.comb += [
            # Connect our transmit multiplexer to the actual UTMI bus.
            tx_multiplexer.output  .attach(self.utmi),

            # Connect up the transmit CRC interface to our UTMI bus.
            data_crc.tx_valid      .eq(tx_multiplexer.output.valid & self.utmi.tx_ready),
            data_crc.tx_data       .eq(tx_multiplexer.output.data),
        ]


        #
        # Device-state management.
        #

        # On a bus reset, clear our address and configuration.
        with m.If(reset_sequencer.bus_reset):
            m.d.usb += [
                address        .eq(0),
                configuration  .eq(0),
            ]


        # Device operating state controls.
        m.d.comb += [
            # Disable our host-mode pulldowns; as we're a device.
            self.utmi.dm_pulldown            .eq(0),
            self.utmi.dp_pulldown            .eq(0),

            # Let our reset sequencer set our USB mode and speed.
            reset_sequencer.low_speed_only   .eq(self.low_speed_only & ~self.always_fs),
            reset_sequencer.full_speed_only  .eq(self.full_speed_only | self.always_fs),
            self.utmi.op_mode                .eq(reset_sequencer.operating_mode),
            self.utmi.xcvr_select            .eq(reset_sequencer.current_speed),
            self.utmi.term_select            .eq(reset_sequencer.termination_select & self.connect),
        ]

        #
        # Frame/microframe state.
        #

        # Handle each new SOF token as we receive them.
        with m.If(token_detector.interface.new_frame):

            # Update our knowledge of the current frame number.
            m.d.usb += self.frame_number.eq(token_detector.interface.frame)

            # Check if we're receiving a new 1ms frame -- which occurs when the new SOF's
            # frame number is different from the previous one's. This will always be the case
            # on full speed links; and will be the case 1/8th of the time on High Speed links.
            m.d.comb += self.new_frame.eq(token_detector.interface.frame != self.frame_number)

            # If this is a new frame, our microframe count should be zero.
            with m.If(self.new_frame):
                m.d.usb += self.microframe_number.eq(0)

            # Otherwise, this SOF indicates a new _microframe_ [USB 2.0: 8.4.3.1].
            with m.Else():
                m.d.usb += self.microframe_number.eq(self.microframe_number + 1)


        #
        # Device-state outputs.
        #
        m.d.comb += [
            self.speed            .eq(reset_sequencer.current_speed),
            self.suspended        .eq(reset_sequencer.suspended),

            self.sof_detected     .eq(token_detector.interface.new_frame),
            self.reset_detected   .eq(reset_sequencer.bus_reset),

            self.tx_activity_led  .eq(tx_multiplexer.output.valid),
            self.rx_activity_led  .eq(self.utmi.rx_valid)
        ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Utilities for managing LUNA gateware toolchains. """

import importlib
import logging
import os
import shutil

def configure_toolchain(platform):
    """ Checks if all the required tools for the Yosys toolchain are available.

        If there are missing tools it will attempt to fall back to YoWASP instead.

        Returns:
            True if a valid toolchain has been configured. Otherwise False.
    """

    if platform.has_required_tools():
        # All good, no further hanky-panky required.
        return True

    # Do we have yowasp available to us?
    logging.info(f"Failed to locate {platform.toolchain} toolchain, trying YoWASP.")
    logging.debug("Checking for required tools:")
    for tool in platform.required_tools:
        logging.debug(f"    {tool}")

    problems = 0

    # Check whether yowasp-yosys is installed:
    try:
        import yowasp_yosys
        logging.debug(f"Found module: {yowasp_yosys.__name__}")
    except Exception as e:
        problems += 1
        logging.warning(e)

    # Check whether yowasp-nextpnr-<target> is installed:
    try:
        nextpnr = next(filter(lambda x: x.startswith("nextpnr-"), platform.required_tools))
        yowasp_nextpnr = importlib.import_module("yowasp_" + nextpnr.replace('-', '_'))
        logging.debug(f"Found module: {yowasp_nextpnr.__name__}")
    except Exception as e:
        problems += 1
        logging.warning(e)

    # Check whether the YoWASP binaries are on the system PATH:
    for tool in platform.required_tools:
        env_var     = tool.replace('-', '_').upper()
        yowasp_tool = "yowasp-" + tool
        path = shutil.which(yowasp_tool)
        if not path:
            problems += 1
            logging.warning(f"'{yowasp_tool}' is not on the system PATH.")
        else:
            logging.debug(f'Setting {env_var}="{yowasp_tool}"')
            os.environ[env_var] = yowasp_tool

    if problems == 0:
        logging.info("YoWASP configured successfully.")
    else:
        logging.info(f"{problems} problems encountered while configuring YoWASP.")
        logging.info(f"You can install it with:")
        logging.info(f"    pip install yowasp-yosys yowasp-{nextpnr}")
        return False

    return True



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Stream generators. """

from amaranth            import Array, Const, DomainRenamer, Elaboratable, Module, Signal
from amaranth.lib.memory import Memory
from .                   import StreamInterface


class ConstantStreamGenerator(Elaboratable):
    """ Gateware that generates stream of constant data.

    Attributes
    ----------
    start: Signal(), input
        Strobe that indicates when the stream should be started.
    done: Signal(), output
        Strobe that pulses high when we're finishing a transmission.

    start_position: Signal(range(len(data)), input
        Specifies the starting position in the constant stream; applied when start() is pulsed.

    max_length: Signal(max_length_width), input
        The maximum length to be sent -in bytes-. Defaults to the length of the stream.
        Only present if the `max_length_width` parameter is provided on creation.
    output_length: Signal(max_length_width), output
        Indicates the actual data length for the stream currently being output.
        Will always be the lesser of our data length and :attr:``max_length``.
        Only present if the `max_length_width` parameter is provided on creation.

    stream: stream_type(), output stream
        The generated stream interface.

    Parameters
    ----------
    constant_data: bytes, or equivalent
        The constant data for the stream to be generated.
        Should be an iterable of integers; or, if data_width is divisible by 8, a bytes-like object.
    domain: string
        The clock domain this generator should belong to. Defaults to 'sync'.
    stream_type: StreamInterface, or subclass
        The type of stream we'll be multiplexing.
    data_width: int, optional
        The width of the constant payload. If not provided; will be taken from the stream's payload width.
    max_length_width: int
        If provided, a `max_length` signal will be present that can limit the total length transmitted.
    data_endianness: little
        If bytes are provided, and our data width is greater
    """


    def __init__(self, constant_data, domain="sync", stream_type=StreamInterface,
            max_length_width=None, data_width=None, data_endianness="little"):

        self._domain           = domain
        self._data             = constant_data
        self._data_length      = len(constant_data)
        self._endianness       = data_endianness
        self._max_length_width = max_length_width

        #
        # I/O port.
        #
        self.start           = Signal()
        self.done            = Signal()

        # If we have a data width, apply it to our stream type; otherwise, use its defaults.
        if data_width:
            self.stream      = stream_type(payload_width=data_width)
            self._data_width = data_width
        else:
            self.stream      = stream_type()
            self._data_width = len(self.stream.data)

        self.start_position = Signal(range(self._data_length))

        # If we have a maximum length width, include it in our I/O port.
        # Otherwise, use a constant.
        if max_length_width:
            self.max_length        = Signal(max_length_width)
            self.output_length     = Signal.like(self.max_length)
        else:
            self.max_length = self._data_length



    def _get_initializer_value(self):
        """ Returns this geneartor's data in a form usable as a ROM initializer.

        Returns
        -------
        initializer_data: interable
            An iterable suitable for use in initializing a ROM.
        valid_bytes_last_word: int
            The number of valid bits that should accompany the last word.

            For example, if we have 32-bit words; and 3 bytes of data, we'd have
            three valid bits on the last word; since the upper 8-bits are meaningless.
        """

        # If we have byte-sized data, Python will implicitly handle things correctly.
        # Return our data unmodified.
        if self._data_width == 8:
            return self._data, len(self.stream.valid)

        # If we don't have a byte-string, return our data without pre-processing.
        if not isinstance(self._data, (bytes, bytearray)):
            return self._data, len(self.stream.valid)

        # If our width isn't evenly divisible by 8, we can't accept bytes.
        if (self._data_width % 8):
            raise ValueError("Can't initialize with bytes unless data_width is divisible by 8!")

        # Figure out how wide each datum will be in bytes.
        datum_width_bytes = self._data_width // 8

        # Otherwise, we'll split it into a list of integers, manually.
        in_data  = bytearray(self._data)
        out_data = []

        while in_data:

            # Extract each datum from our stream...
            datum = in_data[0:datum_width_bytes]
            del in_data[0:datum_width_bytes]

            # ... convert it into an integer ...
            datum = int.from_bytes(datum, byteorder=self._endianness)

            # ... and squish it into our output.
            out_data.append(datum)

        # Figure out how many bytes will be in our last word.
        last_word_bytes = len(self._data) % datum_width_bytes
        if last_word_bytes == 0:
            last_word_bytes = datum_width_bytes

        return out_data, last_word_bytes


    def elaborate(self, platform):
        m = Module()


        #
        # Core ROM.
        #

        # Figure out the shape of our data.
        data_initializer, valid_bits_last_word = self._get_initializer_value()
        data_length = len(data_initializer)

        m.submodules.rom = rom = Memory(shape=self._data_width, depth=data_length, init=data_initializer)
        rom_read_port = rom.read_port()

        if self._max_length_width:
            # Register maximum length, to improve timing.
            max_length = Signal.like(self.max_length)
        else:
            max_length = self.max_length

        # Register that stores our current position in the stream.
        position_in_stream = Signal(range(0, data_length))

        # If we have a maximum length we're enforcing, create a counter for it.
        if self._max_length_width:
            bytes_sent     = Signal(self._max_length_width)
            bytes_per_word = (self._data_width + 7) // 8
        else:
            bytes_sent     = 0
            bytes_per_word = 0


        # Track when we're on the first and last packet.
        on_first_packet = position_in_stream == self.start_position
        on_last_packet  = \
            (position_in_stream          == (data_length - 1)) | \
            (bytes_sent + bytes_per_word >= max_length)


        #
        # Figure out where we should start in our stream.
        #
        start_position = Signal.like(position_in_stream)

        # If our starting position is greater than our data length, use our data length.
        with m.If(self.start_position >= self._data_length):
            m.d.comb += start_position.eq(data_length - 1)

        # Otherwise, use our starting position.
        with m.Else():
            m.d.comb += start_position.eq(self.start_position)


        #
        # Output length field.
        #

        if self._max_length_width:
            # Return our max length or the length of our data, whichever is less.
            with m.If(max_length < self._data_length):
                m.d.comb += self.output_length.eq(max_length)
            with m.Else():
                m.d.comb += self.output_length.eq(self._data_length)




        #
        # Controller.
        #
        with m.FSM(domain=self._domain) as fsm:
            m.d.comb += self.stream.valid.eq(fsm.ongoing('STREAMING'))

            # IDLE -- we're not actively transmitting.
            with m.State('IDLE'):

                # Keep ourselves at the beginning of the stream, but don't yet count.
                m.d.sync += [
                    position_in_stream  .eq(start_position),
                    bytes_sent          .eq(0)
                ]
                m.d.comb += [
                    rom_read_port.addr  .eq(start_position),
                ]

                # Latch the maximum length.
                m.d.sync += [
                    max_length          .eq(self.max_length),
                ]

                # Once the user requests that we start, move to our stream being valid.
                with m.If(self.start & (self.max_length > 0)):
                    m.next = 'STREAMING'


            # STREAMING -- we're actively transmitting data
            with m.State('STREAMING'):
                m.d.comb += [
                    # Always drive the stream from our current memory output...
                    rom_read_port.addr   .eq(position_in_stream),
                    self.stream.payload  .eq(rom_read_port.data),

                    ## ... and base First and Last based on our current position in the stream.
                    self.stream.first    .eq(on_first_packet),
                    self.stream.last     .eq(on_last_packet)
                ]

                # Our ``valid`` flag requires a bunch of special handling, since it could be
                # wider than one bit for streams with multi-byte words; and it could be set
                # by either our max_length limiter or by our data length. This logic is complex,
                # but hopefully actually generates relatively simple hardware.


                # Explicit optimization: if we have a valid length of one, don't bother
                # with all of this logic. This ensures we never degrade speed for trivial cases.
                if len(self.stream.valid) == 1:
                    m.d.comb += self.stream.valid.eq(1)

                # Otherwise, we have more complex logic to deal with.
                else:
                    # If we're on the last packet, we'll apply as many valid bits as we have valid
                    # bytes in our data stream.
                    with m.If(on_last_packet):

                        # If we're not enforcing a max length, always use our leftover bits-per-word.
                        if not self._max_length_width:
                            m.d.comb += self.stream.valid.eq(Const(1).replicate(valid_bits_last_word))

                        # Otherwise, do our complex case.
                        else:
                            # Figure out if we're ending due to the length of data we have, or due to a
                            # maximum-to-send restriction...
                            ending_due_to_data_length = (position_in_stream == (data_length - 1))
                            ending_due_to_max_length  = (bytes_sent + bytes_per_word >= max_length)

                            # ... and figure out the valid bits based us running out of data...
                            valid_due_to_data_length  = Const(1).replicate(valid_bits_last_word)

                            # ... and due to our maximum length. Finding this arithmetically creates
                            # difficult-to-optimize code, and bytes_per_word is going to be small, so
                            # we'll figure this out enumeratively.
                            bytes_left_over         = Signal(range(bytes_per_word + 1))
                            valid_due_to_max_length = Signal.like(self.stream.valid)
                            m.d.comb += bytes_left_over.eq(max_length - bytes_sent)

                            # Generate a case for every possibly number of bytes left over...
                            with m.Switch(bytes_left_over):
                                for i in range(1, bytes_per_word + 1):

                                    # ... with the appropriate amount of valid bits.
                                    with m.Case(i):
                                        m.d.comb += valid_due_to_max_length.eq(Const(1).replicate(i))


                            # Our most complex logic is when both of our end conditions are met; we'll need
                            # to take the lesser of the two validities. AND'ing these will work to accept the
                            # lesser of the two validities.
                            with m.If(ending_due_to_data_length & ending_due_to_max_length):
                                m.d.comb += self.stream.valid.eq(valid_due_to_data_length & valid_due_to_max_length)

                            # If we're ending due to the length of data we have, use our normal logic.
                            with m.Elif(ending_due_to_data_length):
                                m.d.comb += self.stream.valid.eq(valid_due_to_data_length)

                            # Otherwise, we're endign due to our maximum length requirement. We'll apply the
                            # appropriate valid mask.
                            with m.Else():
                                m.d.comb += self.stream.valid.eq(valid_due_to_max_length)


                    # If we're not on our last word, every valid bit should be set.
                    with m.Else():
                        valid_bits = len(self.stream.valid)
                        m.d.comb += self.stream.valid.eq(Const(1).replicate(valid_bits))


                # If the current data byte is accepted, move past it.
                with m.If(self.stream.ready):

                    # If there's still data left to transmit, move forward.
                    with m.If(~on_last_packet):
                        m.d.sync += position_in_stream.eq(position_in_stream + 1)
                        m.d.comb += rom_read_port.addr.eq(position_in_stream + 1)

                        if self._max_length_width:
                            m.d.sync += bytes_sent.eq(bytes_sent + bytes_per_word)

                    # Otherwise, we've finished streaming. Return to IDLE.
                    with m.Else():
                        m.next = 'DONE'

            # DONE -- report our completion; and then return to idle
            with m.State('DONE'):
                m.d.comb += self.done.eq(1)
                m.next = 'IDLE'


        # Convert our sync domain to the domain requested by the user, if necessary.
        if self._domain != "sync":
            m = DomainRenamer({"sync": self._domain})(m)

        return m



class StreamSerializer(Elaboratable):
    """ Gateware that serializes a short Array input onto a stream.

    I/O port:
        I: start        -- Strobe that indicates when the stream should be started.
        O: done         -- Strobe that pulses high when we're finishing a transmission.

        I: data[]       -- The data stream to be sent out. Length is set by the data_length initializer argument.
        I: max_length[] -- The maximum length to be sent. Defaults to the length of the stream.
                           Only present if the `max_length_width` parameter is provided on creation.

        *: stream       -- The generated stream interface.

    """

    def __init__(self, data_length, domain="sync", data_width=8, stream_type=StreamInterface, max_length_width=None):
        """
        Parameters:
            data_length        -- The length of the data to be transmitted.
            domain             -- The clock domain this generator should belong to. Defaults to 'sync'.
            data_width         -- The width of the constant payload
            stream_type        -- The type of stream we'll be multiplexing. Must be a subclass of StreamInterface.
            max_length_width   -- If provided, a `max_length` signal will be present that can limit the total length
                                  transmitted.
        """

        self.domain         = domain
        self.data_width     = data_width
        self.data_length    = data_length

        #
        # I/O port.
        #
        self.start          = Signal()
        self.done           = Signal()

        self.data           = Array(Signal(data_width, name=f"datum_{i}") for i in range(data_length))
        self.stream         = stream_type(payload_width=data_width)

        self.start_position = Signal(range(self.data_length))

        # If we have a maximum length width, include it in our I/O port.
        # Otherwise, use a constant.
        if max_length_width:
            self.max_length = Signal(max_length_width)
        else:
            self.max_length = self.data_length



    def elaborate(self, platform):
        m = Module()

        # Register that stores our current position in the stream.
        position_in_stream = Signal(range(0, self.data_length))
        bytes_sent         = Signal.like(self.max_length)

        # Track when we're on the first and last packet.
        on_first_packet = position_in_stream == self.start_position
        on_last_packet  = \
            (position_in_stream == (self.data_length - 1)) | \
            (bytes_sent         == (self.max_length - 1))

        m.d.comb += [
            # Create first and last based on our stream position.
            self.stream.first    .eq(on_first_packet & self.stream.valid),
            self.stream.last     .eq(on_last_packet  & self.stream.valid)
        ]


        #
        # Figure out where we should start in our stream.
        #
        start_position = Signal.like(position_in_stream)

        # If our starting position is greater than our data length, use our data length.
        with m.If(self.start_position >= self.data_length):
            m.d.comb += start_position.eq(self.data_length - 1)

        # Otherwise, use our starting position.
        with m.Else():
            m.d.comb += start_position.eq(self.start_position)


        #
        # Controller.
        #
        with m.FSM(domain=self.domain) as fsm:
            m.d.comb += self.stream.valid.eq(fsm.ongoing('STREAMING'))

            # IDLE -- we're not actively transmitting.
            with m.State('IDLE'):

                # Keep ourselves at the beginning of the stream, but don't yet count.
                m.d.sync += [
                    position_in_stream  .eq(start_position),
                    bytes_sent          .eq(0),
                ]

                # Once the user requests that we start, move to our stream being valid.
                with m.If(self.start & (self.max_length > 0)):
                    m.next = 'STREAMING'


            # STREAMING -- we're actively transmitting data
            with m.State('STREAMING'):
                m.d.comb += self.stream.payload.eq(self.data[position_in_stream])

                # If the current data byte is accepted, move past it.
                with m.If(self.stream.ready):

                    # If there's still data left to transmit, move forward.
                    with m.If(~on_last_packet):
                        m.d.sync += position_in_stream.eq(position_in_stream + 1)
                        m.d.sync += bytes_sent        .eq(bytes_sent + 1)

                    # Otherwise, we've finished streaming. Return to IDLE.
                    with m.Else():
                        m.next = 'DONE'

            # DONE -- report our completion; and then return to idle
            with m.State('DONE'):
                m.d.comb += self.done.eq(1)
                m.next = 'IDLE'


        # Convert our sync domain to the domain requested by the user, if necessary.
        if self.domain != "sync":
            m = DomainRenamer({"sync": self.domain})(m)

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Import shortcuts for our ready-to-use devices. """

# Create shorthands for the most common parts of the library's usb2 gateware.
from .gateware.usb.devices.acm import USBSerialDevice

__all__ = ['USBSerialDevice']



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Protocol-level Transaction Packet (flow control header) generation. """

from amaranth import *
from amaranth.hdl.rec import DIR_FANOUT, DIR_FANIN

from usb_protocol.types import USBDirection
from usb_protocol.types.superspeed import HeaderPacketType, TransactionPacketSubtype

from ..link.header import HeaderQueue, HeaderPacket

#
# Interfaces.
#

class HandshakeGeneratorInterface(Record):
    """ Interface used by an endpoint to generate transaction packets.

    Attributes
    -----------
    endpoint_number: Signal(7), input to handshake generator
        The endpoint number associated with the transaction packet to be generated.
    retry_required: Signal(), input to handshake generator
        If set, an ACK will be interpreted as a request to re-send the relevant packet.
    next_sequence: Signal(5), input to handshake generator
        Reports the next expected data packet sequence number to the host.

    send_ack: Signal(), input to handshake generator
        Strobe; requests generation of an ACK packet.

    ready: Signal(), output from handshake generator
        Asserted when the handshake generator is ready to accept new signals.
    done: Signal(), output from handshake geneartor.
        Strobe; pulsed when a requested send is complete.
    """

    def __init__(self):
        super().__init__([

            # Parameters.
            ('endpoint_number', 7, DIR_FANIN),
            ('retry_required',  1, DIR_FANIN),
            ('next_sequence',   5, DIR_FANIN),

            # Commands.
            ('send_ack',        1, DIR_FANIN),
            ('send_stall',      1, DIR_FANIN),
            ('send_nrdy',       1, DIR_FANIN),
            ('send_erdy',       1, DIR_FANIN),

            # Status.
            ('ready',           1, DIR_FANOUT),
            ('done',            1, DIR_FANOUT),

        ])


class HandshakeReceiverInterface(Record):
    """ Interface used by an endpoint to generate transaction packets.

    Attributes
    -----------
    endpoint_number: Signal(7), input to handshake generator
        The endpoint number associated with the transaction packet to be generated.
    retry_required: Signal(), input to handshake generator
        If set, an ACK will be interpreted as a request to re-send the relevant packet.
    next_sequence: Signal(5), input to handshake generator
        Reports the next expected data packet sequence number to the host.

    send_ack: Signal(), input to handshake generator
        Strobe; requests generation of an ACK packet.

    ready: Signal(), output from handshake generator
        Asserted when the handshake generator is ready to accept new signals.
    done: Signal(), output from handshake geneartor.
        Strobe; pulsed when a requested send is complete.
    """

    def __init__(self):
        super().__init__([

            # Parameters.
            ('endpoint_number',   7, DIR_FANOUT),
            ('retry_required',    1, DIR_FANOUT),
            ('next_sequence',     5, DIR_FANOUT),
            ('direction',         1, DIR_FANOUT),
            ('host_error',        1, DIR_FANOUT),
            ('packets_pending',   1, DIR_FANOUT),
            ('number_of_packets', 5, DIR_FANOUT),

            # Commands.
            ('status_received',   1, DIR_FANOUT),
            ('ack_received',      1, DIR_FANOUT),
        ])


#
# Transaction header packets.
#

class TransactionHeaderPacket(HeaderPacket):
    DW0_LAYOUT = [
        ('type',                5),
        ('route_string',       20),
        ('device_address',      7),
    ]


class ACKHeaderPacket(TransactionHeaderPacket):
    DW1_LAYOUT = [
        ('subtype',             4),
        ('reserved_0',          2),
        ('retry',               1),
        ('direction',           1),
        ('endpoint_number',     4),
        ('transfer_type',       3),  # Not used in Gen1.
        ('host_error',          1),
        ('number_of_packets',   5),
        ('data_sequence',       5),
        ('reserved_1',          5),
        ('tp_follows',          1),  # Not used in Gen1.
    ]
    DW2_LAYOUT = [
        ('stream_id',          16),
        ('reserved_2',         11),
        ('packets_pending',     1), # Host only.
        ('reserved_3',          4),
    ]


class NRDYHeaderPacket(TransactionHeaderPacket):
    DW1_LAYOUT = [
        ('subtype',             4),
        ('reserved_0',          3),
        ('direction',           1),
        ('endpoint_number',     4),
        ('reserved_1',         20),
    ]


class ERDYHeaderPacket(TransactionHeaderPacket):
    DW1_LAYOUT = [
        ('subtype',             4),
        ('reserved_0',          3),
        ('direction',           1),
        ('endpoint_number',     4),
        ('reserved_1',          4),
        ('number_of_packets',   5),
        ('reserved_2',         11),
    ]


class STALLHeaderPacket(TransactionHeaderPacket):
    DW1_LAYOUT = [
        ('subtype',             4),
        ('reserved_0',          3),
        ('direction',           1),
        ('endpoint_number',     4),
        ('reserved_1',         20),
    ]
    DW2_LAYOUT = [
        ('reserved_2',         27),
        ('packets_pending',     1), # Host only.
        ('reserved_3',          4),
    ]




class StatusHeaderPacket(TransactionHeaderPacket):
    DW1_LAYOUT = [
        ('subtype',             4),
        ('reserved_0',          3),
        ('direction',           1),
        ('endpoint_number',     4),
        ('reserved_1',         20),
    ]



class TransactionPacketGenerator(Elaboratable):
    """ Module responsible for generating Token Packets, for flow control.

    Attributes
    ----------
    header_source: HeaderQueue(), output stream
        The stream that carries any generated header packet requests.
    interface: HandshakeGeneratorInterface(), to/from endpoint
        The control interface for our packet; meant to be carried to various endpoints.

    address: Signal(7), input
        The address associated with the device; used to fill in header packet fields.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.header_source   = HeaderQueue()
        self.interface       = HandshakeGeneratorInterface()

        self.address         = Signal(7)



    def elaborate(self, platform):
        m = Module()

        header_source = self.header_source
        interface     = self.interface

        # Latched versions of each of our arguments.
        endpoint_number = Signal.like(interface.endpoint_number)
        data_error      = Signal.like(interface.retry_required)
        next_sequence   = Signal.like(interface.next_sequence)
        device_address  = Signal.like(self.address)


        def send_packet(response_type, **fields):
            """ Helper that allows us to easily define a packet-send state."""

            # Create a response packet, and mark ourselves as sending it.
            response = response_type()
            m.d.comb += [
                header_source.valid       .eq(1),
                header_source.header      .eq(response),

                response.type             .eq(HeaderPacketType.TRANSACTION),
                response.device_address   .eq(device_address),
                response.endpoint_number  .eq(endpoint_number)
            ]

            # Next, fill in each of the fields:
            for field, value in fields.items():
                m.d.comb += response[field].eq(value)

            with m.If(header_source.ready):
                m.d.comb += interface.done.eq(1)
                m.next = "DISPATCH_REQUESTS"



        with m.FSM(domain="ss"):

            # DISPATCH_REQUESTS -- we're actively waiting for any generation requests that come in;
            # and preparing to handle them.
            with m.State("DISPATCH_REQUESTS"):
                m.d.comb += interface.ready.eq(1)

                # In this state, we'll constantly latch in our parameters.
                m.d.ss += [
                    endpoint_number  .eq(interface.endpoint_number),
                    data_error       .eq(interface.retry_required),
                    next_sequence    .eq(interface.next_sequence),
                    device_address   .eq(self.address)
                ]

                with m.If(interface.send_ack):
                    m.next = "SEND_ACK"
                with m.If(interface.send_stall):
                    m.next = "SEND_STALL"
                with m.If(interface.send_nrdy):
                    m.next = "SEND_NRDY"
                with m.If(interface.send_erdy):
                    m.next = "SEND_NRDY"


            # SEND_ACK -- actively send an ACK packet to our link partner; and wait for that to complete.
            with m.State("SEND_ACK"):
                send_packet(ACKHeaderPacket,
                    subtype           = TransactionPacketSubtype.ACK,
                    direction         = USBDirection.OUT,
                    retry             = data_error,
                    data_sequence     = next_sequence,

                    # TODO: eventually support bursting?
                    number_of_packets = 1,
                )


            # SEND_NRDY -- actively send an NRDY packet to our link partner; and wait for that to complete.
            with m.State("SEND_NRDY"):
                send_packet(NRDYHeaderPacket,
                    subtype           = TransactionPacketSubtype.NRDY,
                    direction         = USBDirection.IN,
                )


            # SEND_NRDY -- actively send an NRDY packet to our link partner; and wait for that to complete.
            with m.State("SEND_ERDY"):
                send_packet(ERDYHeaderPacket,
                    subtype           = TransactionPacketSubtype.ERDY,
                    direction         = USBDirection.IN,

                    # TODO: eventually support bursting?
                    number_of_packets = 1,
                )


            # SEND_STALL -- actively send a STALL packet to our link partner; and wait for that to complete.
            with m.State("SEND_STALL"):
                send_packet(ACKHeaderPacket,
                    subtype           = TransactionPacketSubtype.STALL,
                    direction         = USBDirection.OUT,
                    number_of_packets = 1,
                )


        return m



class TransactionPacketReceiver(Elaboratable):
    """ Module responsible for receiving Transaction Packets from the host.

    Attributes
    ----------
    header_sink: HeaderQueue(), input stream
        The stream that carries received header packets up from the host.
    interface: HandshakeReceiverInterface(), output to endpoint
        Interface that detects transaction packets and reports them to the endpoint.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.header_sink     = HeaderQueue()
        self.interface       = HandshakeReceiverInterface()

        self.address         = Signal(7)


    def elaborate(self, platform):
        m = Module()

        header_sink = self.header_sink
        interface   = self.interface

        # Figure out the vitals for the relevant packet.
        new_packet = header_sink.valid
        is_for_us  = header_sink.get_type() == HeaderPacketType.TRANSACTION

        # Handle any transaction packets.
        with m.If(new_packet & is_for_us):

            # We'll handle the relevant header this cycle; so we can take them off
            # the link layer's hands.
            m.d.comb += header_sink.ready.eq(1)

            # Handle the packet according to its subtype.
            with m.Switch(header_sink.header.dw1[0:4]):

                #
                # STATUS packets -- packets requesting a status stage handshake.
                #
                with m.Case(TransactionPacketSubtype.STATUS):

                    # Handle our packet as a status packet...
                    status_packet = StatusHeaderPacket()
                    m.d.comb += status_packet.eq(header_sink.header)

                    m.d.comb += [
                        # ... fill the fields out in our interface...
                        interface.endpoint_number  .eq(status_packet.endpoint_number),

                        # ... and report the event.
                        interface.status_received  .eq(1)
                    ]


                #
                # ACK packets -- packets that convey transaction status; they both serve as IN
                # tokens and carry any (negative or positive) acknowledgements.
                #
                with m.Case(TransactionPacketSubtype.ACK):

                    # Handle our packet as a status packet...
                    ack_packet = ACKHeaderPacket()
                    m.d.comb += ack_packet.eq(header_sink.header)

                    m.d.comb += [
                        # ... fill the fields out in our interface...
                        interface.endpoint_number    .eq(ack_packet.endpoint_number),
                        interface.retry_required     .eq(ack_packet.retry),
                        interface.next_sequence      .eq(ack_packet.data_sequence),
                        interface.packets_pending    .eq(ack_packet.packets_pending),
                        interface.direction          .eq(ack_packet.direction),
                        interface.host_error         .eq(ack_packet.host_error),
                        interface.number_of_packets  .eq(ack_packet.number_of_packets),

                        # ... and report the event.
                        interface.ack_received       .eq(1)
                    ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test import LunaGatewareTestCase, sync_test_case

from luna.gateware.architecture.car import PHYResetController

class PHYResetControllerTest(LunaGatewareTestCase):
    FRAGMENT_UNDER_TEST = PHYResetController

    def initialize_signals(self):
        yield self.dut.trigger.eq(0)

    @sync_test_case
    def test_power_on_reset(self):

        #
        # After power-on, the PHY should remain in reset for a while.
        #
        yield
        self.assertEqual((yield self.dut.phy_reset), 1)

        yield from self.advance_cycles(30)
        self.assertEqual((yield self.dut.phy_reset), 1)

        yield from self.advance_cycles(60)
        self.assertEqual((yield self.dut.phy_reset), 1)

        #
        # Then, after the relevant reset time, it should resume being unasserted.
        #
        yield from self.advance_cycles(31)
        self.assertEqual((yield self.dut.phy_reset), 0)
        self.assertEqual((yield self.dut.phy_stop),  1)

        yield from self.advance_cycles(120)
        self.assertEqual((yield self.dut.phy_stop),  0)



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test       import LunaUSBGatewareTestCase, LunaSSGatewareTestCase, ss_domain_test_case, usb_domain_test_case

from luna.gateware.stream.generator import ConstantStreamGenerator
from luna.gateware.usb.stream import SuperSpeedStreamInterface

class ConstantStreamGeneratorTest(LunaUSBGatewareTestCase):
    FRAGMENT_UNDER_TEST = ConstantStreamGenerator
    FRAGMENT_ARGUMENTS  = {'constant_data': b"HELLO, WORLD", 'domain': "usb", 'max_length_width': 16}

    @usb_domain_test_case
    def test_basic_transmission(self):
        dut = self.dut

        # Establish a very high max length; so it doesn't apply.
        yield dut.max_length.eq(1000)

        # We shouldn't see a transmission before we request a start.
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.stream.valid), 0)
        self.assertEqual((yield dut.stream.first), 0)
        self.assertEqual((yield dut.stream.last),  0)

        # Once we pulse start, we should see the transmission start,
        # and we should see our first byte of data.
        yield from self.pulse(dut.start)
        self.assertEqual((yield dut.stream.valid),   1)
        self.assertEqual((yield dut.stream.payload), ord('H'))
        self.assertEqual((yield dut.stream.first),   1)

        # That data should remain there until we accept it.
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.stream.valid),   1)
        self.assertEqual((yield dut.stream.payload), ord('H'))

        # Once we indicate that we're accepting data...
        yield dut.stream.ready.eq(1)
        yield

        # ... we should start seeing the remainder of our transmission.
        for i in 'ELLO':
            yield
            self.assertEqual((yield dut.stream.payload), ord(i))
            self.assertEqual((yield dut.stream.first),   0)


        # If we drop the 'accepted', we should still see the next byte...
        yield dut.stream.ready.eq(0)
        yield
        self.assertEqual((yield dut.stream.payload), ord(','))

        # ... but that byte shouldn't be accepted, so we should remain there.
        yield
        self.assertEqual((yield dut.stream.payload), ord(','))

        # If we start accepting data again...
        yield dut.stream.ready.eq(1)
        yield

        # ... we should see the remainder of the stream.
        for i in ' WORLD':
            yield
            self.assertEqual((yield dut.stream.payload), ord(i))


        # On the last byte of data, we should see last = 1.
        self.assertEqual((yield dut.stream.last),   1)

        # After the last datum, we should see valid drop to '0'.
        yield
        self.assertEqual((yield dut.stream.valid), 0)

    @usb_domain_test_case
    def test_basic_start_position(self):
        dut = self.dut

        # Start at position 2
        yield dut.start_position.eq(2)

        # Establish a very high max length; so it doesn't apply.
        yield dut.max_length.eq(1000)

        # We shouldn't see a transmission before we request a start.
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.stream.valid), 0)
        self.assertEqual((yield dut.stream.first), 0)
        self.assertEqual((yield dut.stream.last),  0)

        # Once we pulse start, we should see the transmission start,
        # and we should see our first byte of data.
        yield from self.pulse(dut.start)
        self.assertEqual((yield dut.stream.valid),   1)
        self.assertEqual((yield dut.stream.payload), ord('L'))
        self.assertEqual((yield dut.stream.first),   1)

        # That data should remain there until we accept it.
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.stream.valid),   1)
        self.assertEqual((yield dut.stream.payload), ord('L'))

        # Once we indicate that we're accepting data...
        yield dut.stream.ready.eq(1)
        yield

        # ... we should start seeing the remainder of our transmission.
        for i in 'LO':
            yield
            self.assertEqual((yield dut.stream.payload), ord(i))
            self.assertEqual((yield dut.stream.first),   0)


        # If we drop the 'accepted', we should still see the next byte...
        yield dut.stream.ready.eq(0)
        yield
        self.assertEqual((yield dut.stream.payload), ord(','))

        # ... but that byte shouldn't be accepted, so we should remain there.
        yield
        self.assertEqual((yield dut.stream.payload), ord(','))

        # If we start accepting data again...
        yield dut.stream.ready.eq(1)
        yield

        # ... we should see the remainder of the stream.
        for i in ' WORLD':
            yield
            self.assertEqual((yield dut.stream.payload), ord(i))


        # On the last byte of data, we should see last = 1.
        self.assertEqual((yield dut.stream.last),   1)

        # After the last datum, we should see valid drop to '0'.
        yield
        self.assertEqual((yield dut.stream.valid), 0)

    @usb_domain_test_case
    def test_max_length(self):
        dut = self.dut

        yield dut.stream.ready.eq(1)
        yield dut.max_length.eq(6)

        # Once we pulse start, we should see the transmission start,
        yield from self.pulse(dut.start)

        # ... we should start seeing the remainder of our transmission.
        for i in 'HELLO':
            self.assertEqual((yield dut.stream.payload), ord(i))
            yield


        # On the last byte of data, we should see last = 1.
        self.assertEqual((yield dut.stream.last),   1)

        # After the last datum, we should see valid drop to '0'.
        yield
        self.assertEqual((yield dut.stream.valid), 0)



class ConstantStreamGeneratorWideTest(LunaSSGatewareTestCase):
    FRAGMENT_UNDER_TEST = ConstantStreamGenerator
    FRAGMENT_ARGUMENTS  = dict(
        domain           = "ss",
        constant_data    = b"HELLO WORLD",
        stream_type      = SuperSpeedStreamInterface,
        max_length_width = 16
    )


    @ss_domain_test_case
    def test_basic_transmission(self):
        dut = self.dut

        # Establish a very high max length; so it doesn't apply.
        yield dut.max_length.eq(1000)

        # We shouldn't see a transmission before we request a start.
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.stream.valid), 0)
        self.assertEqual((yield dut.stream.first), 0)
        self.assertEqual((yield dut.stream.last),  0)

        # Once we pulse start, we should see the transmission start,
        # and we should see our first byte of data.
        yield from self.pulse(dut.start)
        self.assertEqual((yield dut.stream.valid),   0b1111)
        self.assertEqual((yield dut.stream.payload), int.from_bytes(b"HELL", byteorder="little"))
        self.assertEqual((yield dut.stream.first),   1)

        # That data should remain there until we accept it.
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.stream.valid),   0b1111)
        self.assertEqual((yield dut.stream.payload), int.from_bytes(b"HELL", byteorder="little"))

        # Once we indicate that we're accepting data...
        yield dut.stream.ready.eq(1)
        yield

        # ... we should start seeing the remainder of our transmission.
        yield
        self.assertEqual((yield dut.stream.valid),   0b1111)
        self.assertEqual((yield dut.stream.payload), int.from_bytes(b"O WO", byteorder="little"))
        self.assertEqual((yield dut.stream.first),   0)


        yield
        self.assertEqual((yield dut.stream.valid),   0b111)
        self.assertEqual((yield dut.stream.payload), int.from_bytes(b"RLD", byteorder="little"))
        self.assertEqual((yield dut.stream.first),   0)


        # On the last byte of data, we should see last = 1.
        self.assertEqual((yield dut.stream.last),   1)

        # After the last datum, we should see valid drop to '0'.
        yield
        self.assertEqual((yield dut.stream.valid), 0)


    @ss_domain_test_case
    def test_max_length_transmission(self):
        dut = self.dut

        # Apply a maximum length of six bytes.
        yield dut.max_length.eq(6)
        yield dut.stream.ready.eq(1)

        # Once we pulse start, we should see the transmission start,
        # and we should see our first byte of data.
        yield from self.pulse(dut.start)
        self.assertEqual((yield dut.stream.valid),   0b1111)
        self.assertEqual((yield dut.stream.payload), int.from_bytes(b"HELL", byteorder="little"))
        self.assertEqual((yield dut.stream.first),   1)

        # We should then see only two bytes of our remainder.
        yield
        self.assertEqual((yield dut.stream.valid),   0b0011)
        self.assertEqual((yield dut.stream.payload), int.from_bytes(b"O WO", byteorder="little"))
        self.assertEqual((yield dut.stream.first),   0)
        self.assertEqual((yield dut.stream.last),    1)


    @ss_domain_test_case
    def test_very_short_max_length(self):
        dut = self.dut

        # Apply a maximum length of six bytes.
        yield dut.max_length.eq(2)

        # Once we pulse start, we should see the transmission start,
        # and we should see our first word of data.
        yield from self.pulse(dut.start)
        self.assertEqual((yield dut.stream.valid),   0b0011)
        self.assertEqual((yield dut.stream.payload), int.from_bytes(b"HELL", byteorder="little"))
        self.assertEqual((yield dut.stream.first),   1)
        self.assertEqual((yield dut.stream.last),    1)

        # Our data should remain there until it's accepted.
        yield dut.stream.ready.eq(1)
        yield
        self.assertEqual((yield dut.stream.valid),   0b0011)
        self.assertEqual((yield dut.stream.payload), int.from_bytes(b"HELL", byteorder="little"))
        self.assertEqual((yield dut.stream.first),   1)
        self.assertEqual((yield dut.stream.last),    1)

        # After acceptance, valid should drop back to false.
        yield
        self.assertEqual((yield dut.stream.valid),   0b0000)




#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" USB Link Commands Transmitter/Receivers """

import functools
import operator

from enum import IntEnum

from amaranth          import *

from .crc              import compute_usb_crc5
from ..physical.coding import SLC, EPF, stream_matches_symbols, get_word_for_symbols
from ...stream         import USBRawSuperSpeedStream


class LinkCommandDetector(Elaboratable):
    """ USB3 Link Command Detector.

    This module detects USB3 link commands as they're received on the bus.

    Attributes
    ----------
    sink: USBRawSuperSpeedStream, input stream
        The (aligned and descrambled) data stream captured from the physical layer.

    command: Signal(4), output
        The link command; including its two-bit class and two-bit type.

    command_class: Signal(2), output
        The link command's class; equivalent to the first two bits of :attr:``command``.
    command_type: Signal(2), output
        The link command's type; equivalent to the second two bits of :attr:``command``.

    subtype: Signal(4), output
        The link command's subtype.

    new_command: signal(), output
        Strobe; indicates that a new link command has been received, and the details of this command
        are ready to be read.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.sink = USBRawSuperSpeedStream()

        # Link command information.
        self.command       = Signal(4)
        self.command_class = Signal(2)
        self.command_type  = Signal(2)
        self.subtype       = Signal(4)

        # Status strobes.
        self.new_command   = Signal()


    def elaborate(self, platform):
        m = Module()

        # Create our ``command_class`` and ``command_type`` aliases,
        # which are always just slices of our command.
        m.d.comb += [
            self.command_class  .eq(self.command[2:4]),
            self.command_type   .eq(self.command[0:2])
        ]

        # Assume we don't have a new command, unless asserted below.
        m.d.ss += self.new_command.eq(0)

        with m.FSM(domain="ss"):

            # WAIT_FOR_LCSTART -- we're currently waiting for LCSTART framing, which indicates
            # that the following word is a link command.
            with m.State("WAIT_FOR_LCSTART"):

                is_lcstart = stream_matches_symbols(self.sink, SLC, SLC, SLC, EPF)
                with m.If(is_lcstart):
                    m.next = "PARSE_COMMAND"


            # PARSE_COMMAND -- our previous data word contained LCSTART; so this word contains our
            # link command. We'll parse / validate it.
            with m.State("PARSE_COMMAND"):

                with m.If(self.sink.valid):

                    link_command_word    = self.sink.data.word_select(0, 16)
                    link_command_replica = self.sink.data.word_select(1, 16)

                    # The payload of a link command contains only data packets; and should never contain
                    # control packets. We'll sanity check this.
                    contains_only_data = (self.sink.ctrl == 0)

                    # A valid two-byte link command word is repeated twice, exactly. [USB3.2r1: 7.2.2.1]
                    # Per the specification, we can only accept commands where both copies match.
                    redundancy_matches = (link_command_word == link_command_replica)

                    # The core ten bits of our link command word are guarded by a CRC-5. We'll only
                    # accept link commands whose CRC matches.
                    crc_matches  = (link_command_word[11:16] == compute_usb_crc5(link_command_word[0:11]))

                    # If we have a word that matches -all- of these criteria, accept it as a new command.
                    with m.If(contains_only_data & redundancy_matches & crc_matches):
                        m.d.ss += [

                            # Copy our fields out of the link command...
                            self.command      .eq(link_command_word[7:11]),
                            self.subtype      .eq(link_command_word[0: 4]),

                            # ... and indicate that we've received a new command
                            self.new_command  .eq(1)
                        ]

                    # No matter the word's validity, we'll move back to waiting for a new command header;
                    # as we can't do anything about invalid commands.
                    m.next = "WAIT_FOR_LCSTART"


        return m



class LinkCommandGenerator(Elaboratable):
    """ USB3 Link Command Generator.

    This module generates link commands on the USB3 bus.

    Attributes
    ----------
    source: USBRawSuperSpeedStream(), output stream
        The data generated in sending our link commands.

    command: Signal(4), output
        The link command; including its two-bit class and two-bit type.
    subtype: Signal(4), output
        The link command's subtype.

    generate: Signal(), input
        Strobe; indicates that a link command should be generated.
    done: Signal(), output
        Indicates that the link command will be complete this cycle; and thus this unit will
        be ready to send another link command next cycle.
    """


    def __init__(self):

        #
        # I/O port
        #
        self.source = USBRawSuperSpeedStream()

        # Link command information.
        self.command    = Signal(4)
        self.subtype    = Signal(4)

        # Control inputs.
        self.generate   = Signal()
        self.done       = Signal()


    def elaborate(self, platform):
        m = Module()

        # Latched versions of our signals guaranteed not to change mid-transmission.
        latched_command = Signal.like(self.command)
        latched_subtype = Signal.like(self.subtype)


        with m.FSM(domain="ss"):

            # IDLE -- we're currently waiting to generate a link command
            with m.State("IDLE"):

                # Once we have a generate command...
                with m.If(self.generate):

                    # ... latch in our command and subtype ...
                    m.d.ss += [
                        latched_command  .eq(self.command),
                        latched_subtype  .eq(self.subtype)
                    ]
                    m.next = "TRANSMIT_HEADER"

            # TRANSMIT_HEADER -- we're starting our link command by transmitting a header.
            with m.State("TRANSMIT_HEADER"):

                # Drive the bus with our header...
                header_data, header_ctrl = get_word_for_symbols(SLC, SLC, SLC, EPF)
                m.d.comb += [
                    self.source.valid  .eq(1),
                    self.source.data   .eq(header_data),
                    self.source.ctrl   .eq(header_ctrl),
                ]

                # ... and keep driving it until it's accepted.
                with m.If(self.source.ready):
                    m.next = "TRANSMIT_COMMAND"


            # TRANSMIT_COMMAND -- we're now ready to send the core of our command.
            with m.State("TRANSMIT_COMMAND"):
                link_command = Signal(16)

                # Drive our command onto the bus...
                m.d.comb += [
                    # First, build our core command...
                    link_command[ 0: 4]      .eq(latched_subtype),
                    link_command[ 4: 7]      .eq(0),  # Reserved.
                    link_command[ 7:11]      .eq(latched_command),
                    link_command[11:16]      .eq(compute_usb_crc5(link_command[0:11])),

                    # ... and then duplicate it as a command onto the output.
                    self.source.valid        .eq(1),
                    self.source.data[ 0:16]  .eq(link_command),
                    self.source.data[16:32]  .eq(link_command),
                    self.source.ctrl         .eq(0)
                ]

                # ... and keep driving it until it's accepted.
                with m.If(self.source.ready):
                    m.d.comb += self.done.eq(1)
                    m.next = "IDLE"


        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Endpoint abstractions for USB3. """

import operator
import functools

from amaranth import *

from .transaction import HandshakeGeneratorInterface, HandshakeReceiverInterface

from ..link.data   import DataHeaderPacket
from ....utils.bus import OneHotMultiplexer
from ...stream     import SuperSpeedStreamInterface

class SuperSpeedEndpointInterface:
    """ Interface that connects a USB3 endpoint module to a USB device.

    Many non-control endpoints won't need to use the latter half of this structure;
    it will be automatically removed by the relevant synthesis tool.

    Attributes
    ----------
    rx: SuperSpeedStreamInterface(), input stream to endpoint
        Receive interface for this endpoint. This stream's ``ready`` signal is ignored.
    rx_header: DataHeaderPacket(), input to endpoint
        The header associated with the packet currently being received.
    rx_complete: Signal(), input to endpoint
        Strobe that indicates that the concluding rx-stream was valid (CRC check passed).
    rx_invalid: Signal(), input to endpoint
        Strobe that indicates that the concluding rx-stream was invalid (CRC check failed).
    rx_new_header: Signal(), input to endpoint
        Strobe; indicates that a new header is available on rx_header.

    tx: SuperSpeedStreamInterface(), output stream from endpoint
        Transmit interface for this endpoint. This stream's ``valid`` must remain high for
        an entire packet; and it must respect the transmitter's ``ready`` signal.
    tx_zlp: Signal(), output from endpoint
        Strobe; when pulsed, triggers sending of a zero-length packet.
    tx_length: Signal(range(1024 + 1)), output from endpoint
        The length of the packet to be transmitted; required for generating its header.
    tx_endpoint_number: Signal(4), output from endpoint
        The endpoint number associated with the active transmission.
    tx_sequence_number: Signal(4), output from endpoint
        The sequence number associated with the active transmission.
    tx_direction: Signal(), output from endpoint
        The direction associated with the active transmission; used for control endpoints.

    active_address: Signal(7), input to endpoint
        Contains the device's current address.
    address_changed: Signal(), output from endpoint.
        Strobe; pulses high when the device's address should be changed.
    new_address: Signal(7), output from endpoint
        When :attr:`address_changed` is high, this field contains the address that should be adopted.

    active_config: Signal(8), input to endpoint
        The configuration number of the active configuration.
    config_changed: Signal(), output from endpoint
        Strobe; pulses high when the device's configuration should be changed.
    new_config: Signal(8)
        When `config_changed` is high, this field contains the configuration that should be applied.
    """

    def __init__(self):

        # Data packet reception.
        self.rx                    = SuperSpeedStreamInterface()
        self.rx_header             = DataHeaderPacket()
        self.rx_complete           = Signal()
        self.rx_invalid            = Signal()

        # Data packet transmission.
        self.tx                    = SuperSpeedStreamInterface()
        self.tx_zlp                = Signal()
        self.tx_length             = Signal(range(1024 + 1))
        self.tx_endpoint_number    = Signal(4)
        self.tx_sequence_number    = Signal(5)
        self.tx_direction          = Signal(init=1)

        # Handshaking / transaction packet exchange.
        self.handshakes_out        = HandshakeGeneratorInterface()
        self.handshakes_in         = HandshakeReceiverInterface()

        # Endpoint state.
        self.ep_reset              = Signal()

        # Typically only used for control endpoints.
        self.active_address        = Signal(7)
        self.address_changed       = Signal()
        self.new_address           = Signal(7)

        self.active_config         = Signal(8)
        self.config_changed        = Signal()
        self.new_config            = Signal(8)



class SuperSpeedEndpointMultiplexer(Elaboratable):
    """ Multiplexes access to the resources shared between multiple endpoint interfaces.

    Interfaces are added using :attr:`add_interface`.

    Attributes
    ----------

    shared: SuperSpeedEndpointInterface
        The post-multiplexer endpoint interface.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.shared = SuperSpeedEndpointInterface()

        #
        # Internals
        #
        self._interfaces = []


    def add_interface(self, interface: SuperSpeedEndpointInterface):
        """ Adds a EndpointInterface to the multiplexer.

        Arbitration is not performed; it's expected only one endpoint will be
        driving the transmit lines at a time.
        """
        self._interfaces.append(interface)


    def _multiplex_signals(self, m, *, when, multiplex):
        """ Helper that creates a simple priority-encoder multiplexer.

        Parmeters
        ---------
        when: str
            The name of the interface signal that indicates that the `multiplex` signals should be
            selected for output. If this signals should be multiplexed, it should be included in `multiplex`.
        multiplex: iterable(str)
            The names of the interface signals to be multiplexed.
        """

        # We're building an if-elif tree; so we should start with an If entry.
        conditional = m.If

        for interface in self._interfaces:
            condition = getattr(interface, when)

            with conditional(condition):

                # Connect up each of our signals.
                for signal_name in multiplex:

                    # Get the actual signals for our input and output...
                    driving_signal = getattr(interface,   signal_name)
                    target_signal  = getattr(self.shared, signal_name)

                    # ... and connect them.
                    m.d.comb += target_signal   .eq(driving_signal)

            # After the first element, all other entries should be created with Elif.
            conditional = m.Elif



    def elaborate(self, platform):
        m = Module()
        shared = self.shared

        #
        # Pass through signals being routed -to- our pre-mux interfaces.
        #
        for interface in self._interfaces:
            m.d.comb += [

                # Rx interface.
                interface.rx                     .tap(shared.rx),
                interface.rx_header              .eq(shared.rx_header),
                interface.rx_complete            .eq(shared.rx_complete),
                interface.rx_invalid             .eq(shared.rx_invalid),

                # Handshake exchange.
                shared.handshakes_in             .connect(interface.handshakes_in),

                # State signals.
                interface.ep_reset               .eq(shared.config_changed),
                interface.active_config          .eq(shared.active_config),
                interface.active_address         .eq(shared.active_address)
            ]

        #
        # Multiplex each of our transmit interfaces.
        #
        for interface in self._interfaces:

            # If the transmit interface is valid, connect it up to our endpoint.
            # The latest assignment will win; so we can treat these all as a parallel 'if's
            # and still get an appropriate priority encoder.
            with m.If(interface.tx.valid.any() | interface.tx_zlp):
                m.d.comb += [
                    shared.tx                  .stream_eq(interface.tx),
                    shared.tx_zlp              .eq(interface.tx_zlp),
                    shared.tx_direction        .eq(interface.tx_direction),
                    shared.tx_endpoint_number  .eq(interface.tx_endpoint_number),
                    shared.tx_sequence_number  .eq(interface.tx_sequence_number),
                    shared.tx_length           .eq(interface.tx_length)
                ]


        #
        # Multiplex each of our handshake-out interfaces.
        #
        for interface in self._interfaces:
            any_generate_signal_asserted = (
                interface.handshakes_out.send_ack   |
                interface.handshakes_out.send_stall
            )

            # If the given interface is trying to send an handshake, connect it up
            # to our shared interface.
            with m.If(any_generate_signal_asserted):
                m.d.comb += shared.handshakes_out.connect(interface.handshakes_out)


        #
        # Multiplex the signals being routed -from- our pre-mux interface.
        #
        self._multiplex_signals(m,
            when='address_changed',
            multiplex=['address_changed', 'new_address']
        )
        self._multiplex_signals(m,
            when='config_changed',
            multiplex=['config_changed', 'new_config']
        )


        return m
